import json
import typing as t
import uuid

import click

import mantik.authentication.auth
import mantik.cli._options as _options
import mantik.cli.main as main
import mantik.cli.utils as utils
import mantik.models.image as image
import mantik.utils.env
import mantik.utils.mantik_api.client
import mantik.utils.mantik_api.models
import mantik.utils.urls

GROUP_NAME = "models"


@main.cli.group(GROUP_NAME)
def cli() -> None:
    """Interaction with models through the mantik api."""


def mlflow_parameters_decoder(ctx, param, value) -> t.Optional[dict]:
    try:
        return json.loads(value) if value is not None else None
    except json.decoder.JSONDecodeError:
        raise TypeError(
            """MLflow parameters are in the wrong format.
            Try: '{"key": "value"}'"""
        )


@cli.command("list")
@_options.PROJECT_ID
def get_all_project_models(project_id: uuid.UUID) -> None:
    """Print details of all trained models in project."""
    models = mantik.utils.mantik_api.models.get_all(
        project_id=project_id,
        token=utils.access_token_from_env_vars(),
    )
    click.echo(models)


@cli.command("get-one")
@_options.PROJECT_ID
@_options.MODEL_ID
def get_model(project_id: uuid.UUID, model_id: uuid.UUID) -> None:
    """Print details of specific trained model."""

    model = mantik.utils.mantik_api.models.get_one(
        project_id=project_id,
        model_id=model_id,
        token=utils.access_token_from_env_vars(),
    )
    click.echo(model)


@cli.command("delete")
@_options.PROJECT_ID
@_options.MODEL_ID
def delete_model(project_id: uuid.UUID, model_id: uuid.UUID) -> None:
    """Delete a trained model."""
    mantik.utils.mantik_api.models.delete(
        project_id=project_id,
        model_id=model_id,
        token=utils.access_token_from_env_vars(),
    )


@cli.command("add")
@_options.PROJECT_ID
@_options.get_name_option(required=True, help_option="Model name.")
@click.option("--uri", type=str, required=False)
@click.option("--location", type=str, required=False)
@_options.get_connection_id(required=False)
@click.option(
    "--mlflow-parameters",
    type=str,
    required=False,
    help="""JSON data as a string.""",
    callback=mlflow_parameters_decoder,
)
@_options.get_run_id(required=False)
def create_model_entry(
    project_id: uuid.UUID,
    name: str,
    uri: str,
    location: str,
    connection_id: t.Optional[uuid.UUID],
    mlflow_parameters: t.Optional[dict],
    run_id: t.Optional[uuid.UUID],
) -> None:
    """Add a new trained model entry."""

    mantik.utils.mantik_api.models.add(
        project_id=project_id,
        new_model_schema=mantik.utils.mantik_api.models.PostPutModel(
            uri=uri,
            location=location,
            connection_id=connection_id,
            mlflow_parameters=mlflow_parameters,
            run_id=run_id,
            name=name,
        ),
        token=utils.access_token_from_env_vars(),
    )


@cli.command("update")
@_options.PROJECT_ID
@_options.MODEL_ID
@_options.get_name_option(required=True, help_option="New Model name.")
@click.option("--uri", type=str, required=True)
@click.option("--location", type=str, required=True)
@_options.get_connection_id(required=False)
@click.option(
    "--mlflow-parameters",
    type=str,
    required=False,
    help="""JSON data as a string.""",
    callback=mlflow_parameters_decoder,
)
@_options.get_run_id(required=False)
def update_model_entry(
    project_id: uuid.UUID,
    model_id: uuid.UUID,
    uri: str,
    name: str,
    location: str,
    connection_id: t.Optional[uuid.UUID],
    mlflow_parameters: t.Optional[dict],
    run_id: t.Optional[uuid.UUID],
) -> None:
    """Update details of existing trained model."""

    mantik.utils.mantik_api.models.update(
        project_id=project_id,
        model_id=model_id,
        updated_model_schema=mantik.utils.mantik_api.models.PostPutModel(
            uri=uri,
            location=location,
            connection_id=connection_id,
            mlflow_parameters=mlflow_parameters,
            run_id=run_id,
            name=name,
        ),
        token=utils.access_token_from_env_vars(),
    )


@cli.command("download")
@_options.PROJECT_ID
@_options.MODEL_ID
@_options.get_target_dir_option(
    help_option="Path to directory "
    "where the zipped tarball image will be downloaded.",
)
@click.option(
    "--image-type",
    type=click.Choice(["docker", "apptainer"], case_sensitive=False),
    show_default=True,
    default="docker",
    help="Type of the image to fetch from the mantik platform.",
)
@click.option(
    "--load/--no-load",
    default=False,
    help="Load the tarball image into docker.",
    show_default=True,
)
def download_model(
    project_id: uuid.UUID,
    model_id: uuid.UUID,
    target_dir: str,
    image_type: str,
    load: bool,
) -> None:
    """Download the containerized version of the model."""
    image.download_container(
        project_id=project_id,
        model_id=model_id,
        target_dir=target_dir,
        image_type=image_type,
        load=load,
    )


@cli.command("build")
@_options.PROJECT_ID
@_options.MODEL_ID
def build_model(
    project_id: uuid.UUID,
    model_id: uuid.UUID,
) -> None:
    """Trigger model containerization."""
    mantik.utils.mantik_api.models.start_build(
        project_id=project_id,
        model_id=model_id,
        token=utils.access_token_from_env_vars(),
    )


@cli.command("build-status")
@_options.PROJECT_ID
@_options.MODEL_ID
def get_build_status(
    project_id: uuid.UUID,
    model_id: uuid.UUID,
) -> None:
    """Print container build status."""
    model = mantik.utils.mantik_api.models.get_one(
        project_id=project_id,
        model_id=model_id,
        token=utils.access_token_from_env_vars(),
    )
    click.echo(
        f"Build Status {model.status!r} for model with ID: {str(model_id)}"
    )



import dataclasses
import typing as t
import uuid

import mantik.utils.mantik_api.client as client


@dataclasses.dataclass
class ExperimentRepository:
    experiment_repository_id: uuid.UUID
    mlflow_experiment_id: str
    name: str
    artifact_location: t.Optional[str]

    @classmethod
    def from_json(cls, json_str):
        return cls(
            experiment_repository_id=uuid.UUID(
                json_str["experimentRepositoryId"]
            ),
            mlflow_experiment_id=json_str["mlflowExperimentId"],
            name=json_str["name"],
            artifact_location=json_str.get("artifactLocation"),
        )

    def to_json(self):
        return {
            "experimentRepositoryId": str(self.experiment_repository_id),
            "mlflowExperimentId": self.mlflow_experiment_id,
            "name": self.name,
            "artifact_location": self.artifact_location,
        }


def get_one(
    experiment_repository_id: uuid.UUID,
    project_id: uuid.UUID,
    token: str,
) -> ExperimentRepository:
    """Retrieves an experiment repository entry from the Mantik API"""

    endpoint = f"/projects/{str(project_id)}/experiments/{str(experiment_repository_id)}"  # noqa
    response = client.send_request_to_mantik_api(
        method="GET", data={}, url_endpoint=endpoint, token=token
    )
    return ExperimentRepository.from_json(response.json())


def get_unique_run_name(
    experiment_repository_id: uuid.UUID,
    project_id: uuid.UUID,
    token: str,
    run_name: str,
) -> str:
    """Retrieves a unique mlflow run name from the Mantik API"""

    endpoint = f"/projects/{str(project_id)}/experiments/{str(experiment_repository_id)}/unique-mlflow-run-name"  # noqa
    response = client.send_request_to_mantik_api(
        method="GET",
        data={},
        url_endpoint=endpoint,
        token=token,
        query_params={"runName": run_name},
    )
    return response.json()



import uuid

import click.testing
import pytest

import mantik.cli.main as main


@pytest.mark.parametrize("unzip", ["--unzip", "--no-unzip"])
def test_download_artifacts_cli(
    mock_authentication,
    tmpdir,
    mock_get_artifacts_url,
    mock_get_url,
    info_caplog,
    unzip,
    zipped_file_name,
):
    runner = click.testing.CliRunner()
    with mock_get_artifacts_url as get_artifacts_url, mock_get_url as get_url:
        result = runner.invoke(
            main.cli,
            [
                "runs",
                "download-artifacts",
                "--project-id",
                str(uuid.uuid4()),
                "--run-id",
                str(uuid.uuid4()),
                "--target-dir",
                tmpdir,
                unzip,
            ],
        )
        get_artifacts_url.assert_called()
        get_url.assert_called()

        assert result.exit_code == 0, result.output



import pathlib

import pytest
from mlflow.projects import _project_spec

import mantik.config.core as core
import mantik.config.exceptions as exceptions
import mantik.utils.credentials as _credentials
import mantik_compute_backend.backend as backend

FILE_PATH = pathlib.Path(__file__).parent

ALL_ENV_VARS = [
    _credentials.UNICORE_USERNAME_ENV_VAR,
    _credentials.UNICORE_PASSWORD_ENV_VAR,
    core.COMPUTE_BUDGET_ACCOUNT_ENV_VAR,
]


def test_ml_project_command_validation():
    entry_point = _project_spec.EntryPoint(
        name="main",
        parameters={
            "optional": {"default": "test", "type": "string"},
        },
        command=(
            "python main.py "
            "{optional} --nproc-per-node=$(${{SLURM_STEP_GPUS: -1}} + 1 )"
        ),
    )
    parameters = {"optional": "whatever"}
    storage_dir = "test-dir"

    backend._create_arguments(
        entry_point=entry_point,
        parameters=parameters,
        storage_dir=storage_dir,
    )


@pytest.mark.parametrize(
    ("entry_point", "user_parameters", "expected"),
    [
        (
            _project_spec.EntryPoint(
                name="main",
                parameters={
                    "required": {"type": "string"},
                    "optional": {"default": "test", "type": "string"},
                },
                command="python main.py {incorrect-parameter} {optional}",
            ),
            {"required": "whatever"},
            "Mismatch between entry point parameter names and parameter "
            "placeholders in the command: ['incorrect-parameter', 'required']. "
            "Please revise the MLproject file for "
            "typos or inconsistencies between parameter names and their "
            "corresponding placeholders in the command.",
        ),
        (
            _project_spec.EntryPoint(
                name="main",
                parameters={
                    "optional": {"default": "test", "type": "string"},
                },
                command=(
                    "python main.py "
                    "{optional} --nproc-per-node=$(${SLURM_STEP_GPUS: -1} + 1 )"
                ),
            ),
            {"optional": "whatever"},
            "Entry point command contains improperly formatted environment "
            "variable expansions: ['${SLURM_STEP_GPUS: -1}']. Ensure "
            "expansions use exactly two opening and closing braces, "
            "e.g. ${{ENV_VAR}}.",
        ),
    ],
    ids=[
        "incorrect parameter in command with optional parameter given",
        "expansion variable given with single curly braces",
    ],
)
def test_ml_project_command_validation_fails(
    entry_point, user_parameters, expected, expect_raise_if_exception
):
    storage_dir = "test-dir"

    with expect_raise_if_exception(
        exceptions.MLprojectFileValidationError()
    ) as e:
        backend._create_arguments(
            entry_point=entry_point,
            parameters=user_parameters,
            storage_dir=storage_dir,
        )
    result = str(e.value)
    assert result == expected


@pytest.mark.parametrize(
    ("entry_point", "user_parameters"),
    [
        (
            _project_spec.EntryPoint(
                name="main",
                parameters={
                    "required": {"type": "string"},
                    "optional": {"default": "test", "type": "string"},
                },
                command="python main.py {required} {optional}",
            ),
            {"required": "whatever"},
        ),
        (
            _project_spec.EntryPoint(
                name="main",
                parameters={
                    "required": {"type": "string"},
                    "optional": {"default": "test", "type": "string"},
                },
                command="python main.py {required} {optional}",
            ),
            {"required": "whatever", "optional": "whatever"},
        ),
    ],
    ids=[
        "required parameter given, optional parameter not given",
        "required and optional parameter given",
    ],
)
def test_validate_parameters_placeholders_entry_point(
    entry_point, user_parameters
):
    backend._validate_parameters_placeholders_entry_point(
        entry_point=entry_point, user_parameters=user_parameters
    )


@pytest.mark.parametrize(
    ("entry_point", "user_parameters", "expected"),
    [
        (
            _project_spec.EntryPoint(
                name="main",
                parameters={
                    "required": {"type": "string"},
                    "optional": {"default": "test", "type": "string"},
                },
                command="python main.py {incorrect-parameter} {optional}",
            ),
            {"required": "whatever"},
            "Mismatch between entry point parameter names and parameter "
            "placeholders in the command: ['incorrect-parameter', 'required']. "
            "Please revise the MLproject file for "
            "typos or inconsistencies between parameter names and their "
            "corresponding placeholders in the command.",
        ),
        (
            _project_spec.EntryPoint(
                name="main",
                parameters={
                    "required": {"type": "string"},
                    "optional": {"default": "test", "type": "string"},
                },
                command="python main.py {required} {optional}",
            ),
            {"incorrect-parameter": "whatever", "optional": "whatever"},
            "Mismatch between entry point parameters and entered parameters: "
            "['incorrect-parameter', 'required']. Please revise the MLproject "
            "file or your submit command for typos or inconsistencies",
        ),
    ],
    ids=[
        "typo in user parameter with optional parameter given",
        "incorrect parameter in command with optional parameter given",
    ],
)
def test_validate_parameters_placeholders_entry_point_fails(
    entry_point, user_parameters, expected, expect_raise_if_exception
):
    with expect_raise_if_exception(
        exceptions.MLprojectFileValidationError()
    ) as e:
        backend._validate_parameters_placeholders_entry_point(
            entry_point=entry_point, user_parameters=user_parameters
        )
    result = str(e.value)
    assert result == expected


def test_environment_variable_expansion_validation():
    broken_entry_point = _project_spec.EntryPoint(
        name="main",
        parameters={},
        command=(
            "python main.py "
            "--nproc-per-node=$(( ${{SLURM_STEP_GPUS: -1}} + 1 ))"
        ),
    )

    backend._validate_env_var_expansion_in_entry_point(
        entry_point=broken_entry_point
    )


@pytest.mark.parametrize(
    ("env_var_expansion", "env_var_expansion_list"),
    [
        ("$(( ${SLURM_STEP_GPUS: -1} + 1 ))", "['${SLURM_STEP_GPUS: -1}']"),
        (
            "$(( ${{{SLURM_STEP_GPUS: -1}}} + 1 ))",
            "['${{{SLURM_STEP_GPUS: -1}}}']",
        ),
        (
            "$(( ${{{SLURM_STEP_GPUS: -1}} + 1 ))",
            "['${{{SLURM_STEP_GPUS: -1}}']",
        ),
    ],
    ids=[
        "Single opening and closing curly braces",
        "Matching number of opening and closing curly braces, "
        "but incorrect number",
        "Not matching numbers of opening and closing curly braces",
    ],
)
def test_environment_variable_expansion_fails(
    env_var_expansion, env_var_expansion_list, expect_raise_if_exception
):
    broken_entry_point = _project_spec.EntryPoint(
        name="main",
        parameters={},
        command=f"python main.py --nproc-per-node={env_var_expansion}",
    )
    expected = (
        "Entry point command contains improperly formatted environment "
        f"variable expansions: {env_var_expansion_list}. Ensure "
        "expansions use exactly two opening and closing braces, "
        "e.g. ${{ENV_VAR}}."
    )
    with expect_raise_if_exception(
        exceptions.MLprojectFileValidationError()
    ) as e:
        backend._validate_env_var_expansion_in_entry_point(
            entry_point=broken_entry_point
        )
    result = str(e.value)
    assert result == expected



"""Interact with trained models in mantik api"""
import dataclasses
import logging
import typing as t
import uuid

import mantik.utils.mantik_api.client as client

logger = logging.getLogger(__name__)

T = t.TypeVar("T")


def optional_cast(_type: T, optional_value: t.Optional[t.Any]) -> t.Optional[T]:
    if optional_value is None:
        return None

    return _type(optional_value)


class InvalidModelError(Exception):
    """The given model payload is invalid"""


@dataclasses.dataclass
class ModelBase:
    name: str


@dataclasses.dataclass
class ModelBaseKeywords:
    uri: t.Optional[str] = None
    location: t.Optional[str] = None
    connection_id: t.Optional[uuid.UUID] = None
    run_id: t.Optional[uuid.UUID] = None
    mlflow_parameters: t.Optional[dict] = None


@dataclasses.dataclass
class GetModelNoKeywords(ModelBase):
    model_id: uuid.UUID
    status: str


@dataclasses.dataclass
class GetModel(ModelBaseKeywords, GetModelNoKeywords):
    @classmethod
    def from_json(cls, json_str):
        return cls(
            model_id=uuid.UUID(json_str["modelId"]),
            uri=json_str["uri"],
            location=json_str["location"],
            connection_id=optional_cast(uuid.UUID, json_str["connectionId"]),
            mlflow_parameters=json_str["mlflowParameters"],
            status=json_str["status"],
            run_id=optional_cast(uuid.UUID, json_str["runId"]),
            name=json_str["name"],
        )


@dataclasses.dataclass
class PostPutModel(ModelBaseKeywords, ModelBase):
    def to_dict(self):
        return {
            "uri": self.uri,
            "location": self.location,
            "connectionId": optional_cast(str, self.connection_id),
            "mlflowParameters": self.mlflow_parameters,
            "runId": optional_cast(str, self.run_id),
            "name": self.name,
        }

    def validate(self):
        if self.run_id and (self.uri and self.location):
            # The precedence is enforced here:
            # https://gitlab.com/mantik-ai/mantik-api/-/blob/main/src/mantik_api/
            # routes/project_routes/model_repository.py?ref_type=heads#L433-436
            logger.warning(
                "If both (run ID) and "
                "(uri and location) are defined the (run ID) takes precedence, "
                "and (uri and location) will be overwritten."
            )
        if (self.run_id is None) and (
            self.uri is None and self.location is None
        ):
            raise InvalidModelError(
                "To add a model either (run ID) or "
                "(uri and location) must be given."
            )


def get_all(
    project_id: uuid.UUID,
    token: str,
) -> t.List[GetModel]:
    """Retrieves trained model entries through the Mantik API"""

    endpoint = f"/projects/{str(project_id)}/models/trained"
    response = client.send_request_to_mantik_api(
        method="GET", data={}, url_endpoint=endpoint, token=token
    )
    return [GetModel.from_json(model) for model in response.json()["models"]]


def get_one(
    project_id: uuid.UUID,
    model_id: uuid.UUID,
    token: str,
) -> GetModel:
    """Retrieves a trained model entry from the Mantik API"""

    endpoint = f"/projects/{str(project_id)}/models/trained/{str(model_id)}"
    response = client.send_request_to_mantik_api(
        method="GET", data={}, url_endpoint=endpoint, token=token
    )
    return GetModel.from_json(response.json())


def delete(
    project_id: uuid.UUID,
    model_id: uuid.UUID,
    token: str,
):
    """Deletes a trained model entry through the Mantik API"""

    endpoint = f"/projects/{str(project_id)}/models/trained/{str(model_id)}"
    client.send_request_to_mantik_api(
        method="DELETE", data={}, url_endpoint=endpoint, token=token
    )
    logger.info(f"Model with ID: {model_id} has been deleted")


def add(
    new_model_schema: PostPutModel,
    project_id: uuid.UUID,
    token: str,
) -> uuid.UUID:
    """Creates a trained model entry through the Mantik API"""

    new_model_schema.validate()
    data = new_model_schema.to_dict()
    endpoint = f"/projects/{str(project_id)}/models/trained"
    response = client.send_request_to_mantik_api(
        method="POST", data=data, url_endpoint=endpoint, token=token
    )
    new_model_id = response.json()["modelId"]
    logger.info(f"A model with ID: {new_model_id} has been created")

    return uuid.UUID(new_model_id)


def update(
    updated_model_schema: PostPutModel,
    project_id: uuid.UUID,
    model_id: uuid.UUID,
    token: str,
):
    """Updates a trained model entry through the Mantik API"""

    data = updated_model_schema.to_dict()
    endpoint = f"/projects/{str(project_id)}/models/trained/{str(model_id)}"
    client.send_request_to_mantik_api(
        method="PUT", data=data, url_endpoint=endpoint, token=token
    )
    logger.info(f"Model with ID: {str(model_id)} has been updated")


def get_image_url(
    project_id: uuid.UUID,
    model_id: uuid.UUID,
    token: str,
) -> str:
    """Retrieves a download url
    for a containerized trained model from the Mantik API"""

    endpoint = (
        f"/projects/{str(project_id)}/models/trained/{str(model_id)}/docker"
    )
    response = client.send_request_to_mantik_api(
        method="GET", data={}, url_endpoint=endpoint, token=token
    )
    return response.json()["url"]


def start_build(
    project_id: uuid.UUID,
    model_id: uuid.UUID,
    token: str,
) -> None:
    """Trigger the build for a containerized trained model
    from the Mantik API"""

    endpoint = (
        f"/projects/{str(project_id)}/models/trained/"
        f"{str(model_id)}/docker/build"
    )
    client.send_request_to_mantik_api(
        method="POST", data={}, url_endpoint=endpoint, token=token
    )
    logger.info(f"Building container for model with ID: {str(model_id)}")



import mantik.utils.credentials
import mantik.utils.env
import mantik.utils.env_vars
import mantik.utils.formatting
import mantik.utils.mantik_api
import mantik.utils.mlflow
import mantik.utils.other
import mantik.utils.urls



import uuid

import mantik.utils.mantik_api as mantik_api


def test_submit_run(mock_mantik_api_request, info_caplog):
    submit_run_data = {
        "name": "run-name",
        "experimentRepositoryId": "3fa85f64-5717-4562-b3fc-2c963f66afa6",
        "codeRepositoryId": "3fa85f64-5717-4562-b3fc-2c963f66afa6",
        "branch": "main",
        "commit": "string",
        "dataRepositoryId": "3fa85f64-5717-4562-b3fc-2c963f66afa6",
        "connectionId": "3fa85f64-5717-4562-b3fc-2c963f66afa6",
        "computeBudgetAccount": "a-budget-account",
        "mlflowMlprojectFilePath": "/path",
        "entryPoint": "main",
        "mlflowParameters": {},
        "backendConfig": {
            "UnicoreApiUrl": "https://zam2125.zam.kfa-juelich.de:9112/JUWELS/rest/core",  # noqa F401
            "Environment": {
                "Apptainer": {"Path": "some/image/path.name", "Type": "local"}
            },
            "Resources": {"Queue": "devel", "Nodes": 1},
        },
    }
    project_id = uuid.uuid4()
    with mock_mantik_api_request(
        method="POST",
        end_point=f"/projects/{project_id}/runs",
        status_code=201,
        json_response={},
        expected_error=204,
    ) as (m, error):
        mantik_api.run.submit_run(
            project_id=project_id,
            submit_run_data=submit_run_data,
            token="test_token",
        )
        assert any(
            "Run has been successfully submitted" in message
            for message in info_caplog.messages
        )
    if error:
        assert any(
            "Call to Mantik API" in message for message in info_caplog.messages
        )


def test_save_run(mock_mantik_api_request, info_caplog):
    project_id = uuid.uuid4()
    with mock_mantik_api_request(
        method="POST",
        end_point=f"/projects/{project_id}/runs",
        status_code=201,
        json_response={},
        expected_error=204,
    ) as (m, error):
        mantik_api.run.save_run(
            project_id=project_id,
            run_data={},
            token="test_token",
        )
        assert any(
            "Run has been successfully saved" in message
            for message in info_caplog.messages
        )


def test_update_run_status(mock_mantik_api_request, info_caplog):
    project_id = uuid.uuid4()
    run_id = uuid.uuid4()
    with mock_mantik_api_request(
        method="PUT",
        end_point=f"/projects/{project_id}/runs/{run_id}/status",
        status_code=200,
        json_response={},
        expected_error=204,
    ) as (m, error):
        mantik_api.run.update_run_status(
            project_id=project_id,
            status="FINISHED",
            token="test_token",
            run_id=run_id,
        )
        assert any(
            "Run status has been successfully updated" in message
            for message in info_caplog.messages
        )


def test_update_logs(mock_mantik_api_request, info_caplog):
    project_id = uuid.uuid4()
    run_id = uuid.uuid4()
    with mock_mantik_api_request(
        method="PUT",
        end_point=f"/projects/{project_id}/runs/{run_id}/logs",
        status_code=200,
        json_response={},
        expected_error=204,
    ) as (m, error):
        mantik_api.run.update_logs(
            project_id=project_id,
            logs="Test \n logs",
            token="test_token",
            run_id=run_id,
        )
        assert any(
            "Run logs has been successfully updated" in message
            for message in info_caplog.messages
        )


def test_get_download_artifact_url(mock_mantik_api_request, info_caplog):
    project_id = uuid.uuid4()
    run_id = uuid.uuid4()
    download_url = "test-url"
    with mock_mantik_api_request(
        method="GET",
        end_point=f"/projects/{project_id}/runs/{run_id}/artifacts",
        status_code=200,
        json_response={"url": download_url},
        expected_error=204,
    ) as (m, error):
        url = mantik_api.run.get_download_artifact_url(
            project_id=project_id,
            token="test_token",
            run_id=run_id,
        )
        assert url == download_url
        assert any(
            "Artifacts' download url successfully fetched" in message
            for message in info_caplog.messages
        )



import os

import pytest

import mantik.utils.env as env


@pytest.mark.parametrize(
    ("name", "expected"),
    [
        ("TEST", "test"),
        ("TEST_NOT_SET", NameError()),
    ],
)
def test_get_required_env_var(expect_raise_if_exception, name, expected):
    if not isinstance(expected, Exception):
        os.environ[name] = "test"

    with expect_raise_if_exception(expected):
        result = env.get_required_env_var(name)
        os.unsetenv(name)

        assert result == expected


def test_set_env_vars():
    name = "TEST_ENV_VAR"
    value = "test-value"
    env_vars = {
        name: value,
    }

    env.set_env_vars(env_vars)

    assert os.getenv(name) == value

    os.unsetenv(name)



import logging
import pathlib
import typing as t
import uuid

import click

import mantik.cli._options as main_options
import mantik.cli.runs._options as _options
import mantik.cli.runs.runs as runs
import mantik.cli.utils as utils
import mantik.config.core as core
import mantik.runs.remote
import mantik.runs.schemas

logger = logging.getLogger(__name__)


@runs.cli.command("submit")
@_options.MLPROJECT_PATH
@main_options.get_name_option(required=True, help_option="Name of the Run.")
@_options.ENTRY_POINT
@click.option(
    "--backend-config",
    type=click.Path(dir_okay=False, path_type=pathlib.Path),
    required=True,
    help="Relative or absolute path to backend config file.",
)
@main_options.PROJECT_ID
@main_options.DATA_REPOSITORY_ID_OPTIONAL
@click.option(
    "--data-commit",
    required=False,
    type=str,
    help="Commit hash of data repo to checkout. Takes precedence over data-branch",  # noqa E501
    envvar=mantik.utils.env_vars.DATA_REPOSITORY_COMMIT_ENV_VAR,
    default=None,
)
@click.option(
    "--data-branch",
    required=False,
    type=str,
    help="Branch of data repo to checkout. Defaults main branch.",
    envvar=mantik.utils.env_vars.DATA_REPOSITORY_BRANCH_ENV_VAR,
    default=None,
)
@main_options.EXPERIMENT_REPOSITORY_ID
@main_options.CODE_REPOSITORY_ID
@_options.BRANCH
@_options.COMMIT
@click.option(
    "--compute-budget-account",
    type=str,
    default=None,
    help=f"""Name of your Compute Budget Account on HPC

        If not specified, it is inferred from the environment variable
        {core.COMPUTE_BUDGET_ACCOUNT_ENV_VAR}.

    """,
    envvar=core.COMPUTE_BUDGET_ACCOUNT_ENV_VAR,
)
@_options.PARAMETER
@main_options.VERBOSE
@main_options.get_connection_id(required=True)
def run_project(
    name: str,
    mlproject_path: pathlib.Path,
    entry_point: str,
    backend_config: pathlib.Path,
    parameter: t.List[str],
    verbose: bool,  # noqa
    project_id: t.Optional[uuid.UUID],
    data_repository_id: t.Optional[uuid.UUID],
    data_branch: t.Optional[str],
    data_commit: t.Optional[str],
    experiment_repository_id: t.Optional[uuid.UUID],
    code_repository_id: t.Optional[uuid.UUID],
    branch: t.Optional[str],
    commit: t.Optional[str],
    compute_budget_account: t.Optional[str],
    connection_id: t.Optional[uuid.UUID],
) -> None:
    """Submit an MLflow project as a run to the Mantik Compute Backend.

    Note that `MLPROJECT_PATH` is the relative path to the MLflow project
    folder with your Code Repository as root.

    Remember that when you submit a run, the code is retrieved from your
    remote Git Code Repository. So make sure to commit and push your
    changes before submitting a run! The only file read from your local
    system is the backend config.

    To find the respective required IDs make sure to check Mantik's UI

    """
    _options.check_commit_or_branch(branch=branch, commit=commit, logger=logger)

    logger.debug("Parsing MLflow entry point parameters")
    parameters = utils.dict_from_list(parameter)

    response = mantik.runs.remote.submit_run(
        project_id=project_id,
        name=name,
        experiment_repository_id=experiment_repository_id,
        data_repository_id=data_repository_id,
        data_branch=data_branch,
        data_commit=data_commit,
        code_repository_id=code_repository_id,
        branch=branch,
        commit=commit,
        connection_id=connection_id,
        compute_budget_account=compute_budget_account,
        mlflow_mlproject_file_path=mlproject_path.as_posix(),
        entry_point=entry_point,
        mlflow_parameters=parameters,
        backend_config=backend_config,
    )
    click.echo(response.content)



import contextlib
import os
import platform
import socket
import sys
import typing as t

import GPUtil
import psutil


@contextlib.contextmanager
def temp_chdir(path: str):
    _old_cwd = os.getcwd()
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(_old_cwd)


def construct_git_clone_uri(
    uri: str, git_access_token: str, platform: str
) -> str:
    # How github expects the URI to look like for cloning
    if platform == "GitLab":
        uri = uri.replace("https://", f"https://oauth:{git_access_token}@")
    else:  # GitHub
        uri = uri.replace("https://", f"https://{git_access_token}@")
    return uri


def get_os():
    return platform.system()


def get_cpu_cores():
    return psutil.cpu_count(logical=True)


def get_gpu_info() -> t.List[t.Dict]:
    """Retrieve GPU info.

    Currently only Nvidia GPUs are supported for bookkeeping
    """

    return [
        {
            "name": gpu.name,
            "id": str(gpu.id),
            "driver": gpu.driver,
            "totalMemory": str(gpu.memoryTotal),
        }
        for gpu in GPUtil.getGPUs()
    ]


def get_hostname():
    return socket.gethostname()


def get_memory_gb():
    return f"{psutil.virtual_memory().total / (1024**3):.2f} GB"


def get_platform():
    return platform.platform()


def get_processor():
    return platform.processor()


def get_python_version():
    return platform.python_version()


def get_python_executable():
    return sys.executable



import os

import mantik_compute_backend.utils.temp_dir as temp_dir


def test_use_temp_dir():
    @temp_dir.use_temp_dir
    def return_temp_dir_name(temp_dir_name=None):
        assert os.path.exists(temp_dir_name)
        return temp_dir_name

    assert not os.path.exists(return_temp_dir_name())



import argparse

import pytest

import mantik.cli.utils


@pytest.mark.parametrize(
    ("string_list", "expected_dict"),
    [
        (
            [
                "n_components=3",
                "n_components2=2.7",
                "data='/opt/data/temperature_level_128'",
                "url==========",
            ],
            {
                "n_components": 3,
                "n_components2": 2.7,
                "data": "/opt/data/temperature_level_128",
                "url": "=========",
            },
        ),
    ],
)
def test_dict_from_list(string_list, expected_dict):
    assert expected_dict == mantik.cli.utils.dict_from_list(string_list)


@pytest.mark.parametrize(
    ("parameter_sting", "expected"),
    [
        ("n_components=3", ("n_components", 3)),
        ("n_components=2.7", ("n_components", 2.7)),
        (
            "data='/opt/data/temperature_level_128_daily_averages_2020.nc'",
            ("data", "/opt/data/temperature_level_128_daily_averages_2020.nc"),
        ),
        (
            'data="/opt/data/temperature_level_128_daily_averages_2020.nc"',
            ("data", "/opt/data/temperature_level_128_daily_averages_2020.nc"),
        ),
        (
            "data=/opt/data/temperature_level_128_daily_averages_2020.nc",
            ("data", "/opt/data/temperature_level_128_daily_averages_2020.nc"),
        ),
        ("url==========", ("url", "=========")),
    ],
)
def test_parse_parameter_from_string(parameter_sting, expected):
    assert expected == mantik.cli.utils._parse_parameter_from_string(
        parameter_sting
    )


@pytest.mark.parametrize(
    ("string", "expected"),
    [
        ("3", 3),
        ("2.7", 2.7),
        (
            "/opt/data/temperature_level_128_daily_averages_2020.nc",
            "/opt/data/temperature_level_128_daily_averages_2020.nc",
        ),
    ],
)
def test_parse_value(string, expected):
    assert expected == mantik.cli.utils._parse_value(string)


def test_parse_value_try_injection():
    injection = "_parse_parameter_from_string('a=2')"
    with pytest.raises(argparse.ArgumentTypeError) as parse_error:
        mantik.cli.utils._parse_value(injection)
    assert f"Unable to parse {injection}" in str(parse_error.value)



import dataclasses
import typing as t

import mantik.config._base as _base
import mantik.config._utils as _utils


@dataclasses.dataclass
class Resources(_base.ConfigObject):
    """The computing resources that will be requested ."""

    queue: str
    runtime: t.Optional[str] = None
    nodes: t.Optional[int] = None
    total_cpus: t.Optional[int] = None
    cpus_per_node: t.Optional[int] = None
    gpus_per_node: t.Optional[int] = None
    memory_per_node: t.Optional[str] = None
    reservation: t.Optional[str] = None
    node_constraints: t.Optional[str] = None
    qos: t.Optional[str] = None

    @classmethod
    def _from_dict(cls, config: t.Dict) -> "Resources":
        queue = _utils.get_required_config_value(
            name="Queue",
            value_type=str,
            config=config,
        )
        runtime = _utils.get_optional_config_value(
            name="Runtime",
            value_type=str,
            config=config,
        )
        nodes = _utils.get_optional_config_value(
            name="Nodes",
            value_type=int,
            config=config,
        )
        total_cpus = _utils.get_optional_config_value(
            name="TotalCPUs",
            value_type=int,
            config=config,
        )
        cpus_per_node = _utils.get_optional_config_value(
            name="CPUsPerNode",
            value_type=int,
            config=config,
        )
        gpus_per_node = _utils.get_optional_config_value(
            name="GPUsPerNode",
            value_type=int,
            config=config,
        )
        memory_per_node = _utils.get_optional_config_value(
            name="MemoryPerNode",
            value_type=str,
            config=config,
        )
        reservation = _utils.get_optional_config_value(
            name="Reservation",
            value_type=str,
            config=config,
        )
        node_constraints = _utils.get_optional_config_value(
            name="NodeConstraints",
            value_type=str,
            config=config,
        )
        qos = _utils.get_optional_config_value(
            name="QoS",
            value_type=str,
            config=config,
        )
        return cls(
            queue=queue,
            runtime=runtime,
            nodes=nodes,
            total_cpus=total_cpus,
            cpus_per_node=cpus_per_node,
            gpus_per_node=gpus_per_node,
            memory_per_node=memory_per_node,
            reservation=reservation,
            node_constraints=node_constraints,
            qos=qos,
        )

    def _to_dict(self) -> t.Dict:
        return {
            "Runtime": self.runtime,
            "Queue": self.queue,
            "Nodes": self.nodes,
            "TotalCPUs": self.total_cpus,
            "CPUsPerNode": self.cpus_per_node,
            "GPUS": self.gpus_per_node,
            "MemoryPerNode": self.memory_per_node,
            "Reservation": self.reservation,
            "NodeConstraints": self.node_constraints,
            "QoS": self.qos,
        }

    def to_slurm_batch_script(self):
        sbatch_options_list = [f"#SBATCH --partition={self.queue}"]
        if self.runtime is not None:
            sbatch_options_list.append(
                f"#SBATCH --time={_convert_to_slurm_time_format(self.runtime)}"
            )
        if self.nodes is not None:
            sbatch_options_list.append(f"#SBATCH --nodes={self.nodes}")
        if self.total_cpus is not None:
            sbatch_options_list.append(f"#SBATCH --ntasks={self.total_cpus}")
        if self.cpus_per_node is not None:
            sbatch_options_list.append(
                f"#SBATCH --ntasks-per-node={self.cpus_per_node}"
            )
        if self.gpus_per_node is not None:
            sbatch_options_list.append(
                f"#SBATCH --gpus-per-node={self.gpus_per_node}"
            )
        if self.memory_per_node is not None and int(self.memory_per_node) >= 0:
            sbatch_options_list.append(f"#SBATCH --mem={self.memory_per_node}")
        if self.reservation is not None:
            sbatch_options_list.append(
                f"#SBATCH --reservation={self.reservation}"
            )
        if self.node_constraints is not None:
            sbatch_options_list.append(
                f"#SBATCH --constraint={self.node_constraints}"
            )
        if self.qos is not None:
            sbatch_options_list.append(f"#SBATCH --qos={self.qos}")
        return "\n".join(sbatch_options_list)


def _convert_to_slurm_time_format(input_time: str) -> str:
    """Convert a given time string to the expected format of the `--time`
    option of Slurm.

    UNICORE allows to define time using strings with `d`, `h`, `m`, `s` units.
    Hence, we must convert them here to be consistent with UNICORE.

    """

    if input_time == "0":
        return input_time

    units = {"s": 1, "m": 60, "h": 3600, "d": 86400}

    number, unit = input_time[:-1], input_time[-1]

    # Ensure that given unit (last string character)
    # is a supported time unit.
    if unit not in units:
        time_units = ", ".join(units.keys())
        raise ValueError(
            f"Unknown runtime unit '{unit}'. "
            f"Supported time units: {time_units}."
        )

    # Verify given number.
    # It may be a combination of multiple time units,
    # which is not allowed (UNICORE doesn't support this).
    # E.g. 1h30m is not allowed.
    try:
        value = int(number)
    except ValueError as e:
        raise ValueError(
            "Invalid runtime value. "
            "Only single unit allowed (30m, 12h, 1d, etc.)"
        ) from e

    # Convert the input time to total seconds
    seconds = value * units.get(unit)

    # Convert seconds to target format
    days, remainder = divmod(seconds, 86400)
    hours, remainder = divmod(remainder, 3600)
    minutes, seconds = divmod(remainder, 60)

    return f"{days}-{hours:02}:{minutes:02}:{seconds:02}"



import pathlib
import shutil
import unittest.mock

import fs
import pytest

import mantik.remote_file_service.unicore_file_service as unicore_fs


LOCAL_WORKING_DIRECTORY = pathlib.Path(__file__).parent / "workspace"

SAMPLE_FOLDER = "sample-directory"
SAMPLE_FILENAME = "hello-world.txt"
SAMPLE_FILE_FILE_PATH = f"{SAMPLE_FOLDER}/{SAMPLE_FILENAME}"

USER = "demouser"

AUTH_URL = "http://test-unicore"


@pytest.fixture
def temp_testing_folder_path(tmp_path) -> str:
    # Copy the source folder to the temporary directory
    shutil.copytree(
        f"{LOCAL_WORKING_DIRECTORY}/{SAMPLE_FOLDER}",
        tmp_path / "local" / SAMPLE_FOLDER,
    )

    yield str(tmp_path)


@pytest.fixture()
def unicore_fs_client(
    temp_testing_folder_path, requests_mock
) -> unicore_fs.UnicoreFileService:
    requests_mock.post(
        url=f"{AUTH_URL}",
        json={
            "serverHost": "http://uftp-host",
            "serverPort": 9000,
            "secret": "secret123",
        },
    )
    client = unicore_fs.UnicoreFileService(
        local_base_path=temp_testing_folder_path + "/local",
        username=USER,
        password="test",
        auth_url="http://test-unicore",
    )

    return client


@pytest.fixture()
def mock_get_remote_fs(temp_testing_folder_path):
    fake_remote_path = pathlib.Path(temp_testing_folder_path) / "remote"
    fake_remote_path.mkdir()
    shutil.copytree(
        f"{LOCAL_WORKING_DIRECTORY}/{SAMPLE_FOLDER}",
        fake_remote_path / SAMPLE_FOLDER,
    )
    with unittest.mock.patch(
        "mantik.remote_file_service.unicore_file_service.UnicoreFileService._get_remote_fs",  # noqa E501
        return_value=fs.open_fs(fake_remote_path.as_posix()),
    ) as patch:
        yield patch


def test_list_directory(unicore_fs_client, mock_get_remote_fs) -> None:
    local_list_dir = unicore_fs_client.list_directory(SAMPLE_FOLDER)
    assert len(local_list_dir) == 1
    assert local_list_dir[0].path == SAMPLE_FILE_FILE_PATH
    remote_list_dir = unicore_fs_client.list_directory(
        f"remote://{SAMPLE_FOLDER}"
    )
    assert len(remote_list_dir) == 1
    assert (
        remote_list_dir[0].path == f"remote:/{SAMPLE_FOLDER}/{SAMPLE_FILENAME}"
    )


def test_copy_file(unicore_fs_client, mock_get_remote_fs) -> None:
    assert not unicore_fs_client.exists(f"remote://{SAMPLE_FILENAME}")

    unicore_fs_client.copy_file(
        f"{SAMPLE_FILE_FILE_PATH}",
        f"remote://{SAMPLE_FILENAME}",
    )
    assert unicore_fs_client.exists(f"remote://{SAMPLE_FILENAME}")


def test_remove_file(unicore_fs_client, mock_get_remote_fs) -> None:
    assert unicore_fs_client.exists(
        f"remote://{SAMPLE_FOLDER}/{SAMPLE_FILENAME}"
    )
    unicore_fs_client.remove_file(f"remote://{SAMPLE_FOLDER}/{SAMPLE_FILENAME}")
    assert not unicore_fs_client.exists(
        f"remote://{SAMPLE_FOLDER}/{SAMPLE_FILENAME}"
    )

    #  local
    assert unicore_fs_client.exists(SAMPLE_FILE_FILE_PATH)
    unicore_fs_client.remove_file(SAMPLE_FILE_FILE_PATH)
    assert not unicore_fs_client.exists(SAMPLE_FILE_FILE_PATH)


def test_copy_directory(unicore_fs_client, mock_get_remote_fs) -> None:
    unicore_fs_client.remove_directory(f"remote://{SAMPLE_FOLDER}")
    assert not unicore_fs_client.exists(f"remote://{SAMPLE_FOLDER}")

    unicore_fs_client.copy_directory(SAMPLE_FOLDER, f"remote://{SAMPLE_FOLDER}")
    assert unicore_fs_client.exists(f"remote://{SAMPLE_FOLDER}")

    assert len(
        unicore_fs_client.list_directory(f"remote://{SAMPLE_FOLDER}")
    ) == len(unicore_fs_client.list_directory(SAMPLE_FOLDER))


def test_remove_directory(unicore_fs_client, mock_get_remote_fs) -> None:
    assert unicore_fs_client.exists(f"remote://{SAMPLE_FOLDER}")
    unicore_fs_client.remove_directory(f"remote://{SAMPLE_FOLDER}")
    assert not unicore_fs_client.exists(f"remote://{SAMPLE_FOLDER}")

    #  local
    assert unicore_fs_client.exists(SAMPLE_FOLDER)
    unicore_fs_client.remove_directory(SAMPLE_FOLDER)
    assert not unicore_fs_client.exists(SAMPLE_FOLDER)


def test_create_directory(unicore_fs_client, mock_get_remote_fs) -> None:
    unicore_fs_client.create_directory("remote://new-folder")
    assert unicore_fs_client.exists("remote://new-folder")


def test_user(unicore_fs_client) -> None:
    assert unicore_fs_client.user == USER



import mantik.config.core
import mantik.config.environment
import mantik.config.exceptions
import mantik.config.executable
import mantik.config.firecrest
import mantik.config.read
import mantik.config.resources
import mantik.config.validate



import pytest

import mantik.utils.urls as urls


@pytest.mark.parametrize(
    ("url", "expected"),
    [
        (
            "http://test-url.com/path//with/double//slashes",
            "https://test-url.com/path/with/double/slashes",
        ),
        # Test case: also works with 3 slashes
        (
            "http://test-url.com///path//with/double//slashes",
            "https://test-url.com/path/with/double/slashes",
        ),
    ],
)
def test_ensure_https_and_remove_double_slashes_from_path(url, expected):
    result = urls.remove_double_slashes_from_path(url)

    assert result == expected


@pytest.mark.parametrize(
    ("url", "expected"),
    [
        (
            "http://test.domain.de",
            "https://replaced.domain.de",
        ),
        (
            "https://test.domain.de",
            "https://replaced.domain.de",
        ),
        (
            "http://www.test.domain.de",
            "https://www.replaced.domain.de",
        ),
        (
            "https://www.test.domain.de",
            "https://www.replaced.domain.de",
        ),
        (
            "https://www.test.another.domain.de",
            "https://www.replaced.another.domain.de",
        ),
    ],
)
def test_replace_first_subdomain(url, expected):
    result = urls.replace_first_subdomain(url, replace_with="replaced")

    assert result == expected


@pytest.mark.parametrize(
    ("target_dir", "url", "filetype", "expected"),
    [
        (
            "./something",
            "./something",
            ".tar.gz",
            "./something/something.tar.gz",
        ),
        (
            "./.././this/path/",
            "https://cloud.mantik.ai/image.tar.gz?signature-is-very-long",
            ".tar.gz",
            "./.././this/path/image.tar.gz",
        ),
        (
            "/this/path",
            "https://cloud.mantik.ai/image",
            ".tar.gz",
            "/this/path/image.tar.gz",
        ),
        (
            "/this/path",
            "https://cloud.mantik.ai/image.zip",
            ".zip",
            "/this/path/image.zip",
        ),
    ],
)
def test_get_image_path(target_dir, url, filetype, expected):
    assert (
        urls.get_local_path_from_url(
            target_dir=target_dir, url=url, filetype=filetype
        )
        == expected
    )



import pathlib
import typing as t

import mantik.config.core as core
import mantik.config.executable as executable


def get_files_to_upload(
    project_dir: pathlib.Path,
    config: core.Config,
) -> t.List[pathlib.Path]:
    """Get all files to be uploaded via UNICORE.

    Notes
    -----
    Since MLflow docker backend mounts the MLflow project directory, all
    directory contents are uploaded here as well.

    """
    exclude_list = config.files_to_exclude
    files = _recursively_list_files_in_directory(project_dir, exclude_list)
    if (
        config.execution_environment_given()
        and isinstance(config.environment.execution, executable.Apptainer)
        and config.environment.execution.is_local
    ):
        files.append(
            config.environment.execution.path_as_absolute_to(project_dir)
        )
    return list(set(files))


def _recursively_list_files_in_directory(
    project_dir: pathlib.Path, exclude: t.List[str]
) -> t.List[pathlib.Path]:
    files = [
        path
        for path in project_dir.rglob("*")
        if path.is_file()
        and not file_matches_exclude_pattern(
            path, project_dir=project_dir, exclude=exclude
        )
    ]
    return files


def file_matches_exclude_pattern(
    path: pathlib.Path, project_dir: pathlib.Path, exclude: t.List[str]
) -> bool:
    return _file_matches_exclude_entry(
        path, exclude=exclude
    ) or _file_in_excluded_subdirectory(
        path.relative_to(project_dir), exclude=exclude
    )


def _file_matches_exclude_entry(path: pathlib.Path, exclude: t.List) -> bool:
    return any(
        path.name in _files_matching_pattern(path.parent, pattern)
        for pattern in exclude
    )


def _files_matching_pattern(path: pathlib.Path, pattern: str) -> t.List[str]:
    return [file.name for file in path.glob(pattern)]


def _file_in_excluded_subdirectory(
    element_relative_path: pathlib.Path, exclude: t.List[str]
) -> bool:
    return any(f"{part}/" in exclude for part in element_relative_path.parts)



import datetime
import logging
import typing as t
import uuid

import mantik.remote_file_service.abstract_file_service as abc_fs
import mantik.remote_file_service.unicore_file_service as uc_fs

logger = logging.getLogger()

fake_meta = abc_fs.FileMeta(
    owner="FAKE-OWNER", mode="XXX", last_changed=datetime.datetime(2022, 2, 1)
)

fake_file = abc_fs.File(
    path="remote:/fake-dir/fake.file",
    metadata=fake_meta,
    is_remote=True,
)


fake_directory = abc_fs.Directory(
    path="remote:/fake-dir",
    metadata=fake_meta,
    children=[
        abc_fs.Directory(
            path="remote:/fake-dir/fake-child-dir",
            metadata=fake_meta,
            is_remote=True,
            children=[],
        ),
        fake_file,
    ],
    is_remote=True,
)


class FakeUnicoreFileService(uc_fs.UnicoreFileService):
    def __init__(self):
        pass

    def list_directory(
        self, target
    ) -> t.List[t.Union[abc_fs.Directory, abc_fs.File]]:
        logger.debug(f"list_directory {target}")
        return fake_directory.children

    def create_directory(self, target) -> abc_fs.Directory:
        logger.debug(f"create_directory {target}")
        return fake_directory

    def remove_directory(self, target) -> None:
        logger.debug(f"remove_directory {target}")
        pass

    def copy_directory(
        self,
        source,
        target,
    ) -> abc_fs.Directory:
        logger.debug(f"copy_directory {source} {target}")
        return fake_directory

    def create_file_if_not_exists(self, target) -> abc_fs.File:
        logger.debug(f"create_file_if_not_exists {target}")
        return fake_file

    def remove_file(self, target) -> None:
        logger.debug(f"remove_file {target}")

    def copy_file(self, source, target) -> abc_fs.File:
        logger.debug(f"copy_file {source} {target}")
        return fake_file

    def exists(self, target) -> bool:
        return True

    @property
    def user(self) -> str:
        return "FAKE-USER"

    def change_permissions(self, target, new_permissions) -> None:
        pass

    @classmethod
    def from_env(cls, connection_id: t.Optional[uuid.UUID] = None):
        pass



import mantik.testing.cognito
import mantik.testing.config
import mantik.testing.env
import mantik.testing.firecrest
import mantik.testing.mlflow_server
import mantik.testing.pyunicore
import mantik.testing.remote_file_service
import mantik.testing.requests
import mantik.testing.stress
import mantik.testing.token



import dataclasses
import typing as t
import uuid

import mantik.utils.mantik_api.client as client


@dataclasses.dataclass
class CodeRepository:
    code_repository_id: uuid.UUID
    code_repository_name: str
    uri: str
    connection_id: t.Optional[str]
    platform: str

    @classmethod
    def from_json(cls, json_str):
        return cls(
            code_repository_id=uuid.UUID(json_str["codeRepositoryId"]),
            code_repository_name=json_str["codeRepositoryName"],
            uri=json_str["uri"],
            connection_id=json_str["connectionId"],
            platform=json_str["platform"],
        )

    def to_json(self):
        return {
            "codeRepositoryId": str(self.code_repository_id),
            "codeRepositoryName": self.code_repository_name,
            "uri": self.uri,
            "connectionId": self.connection_id,
            "platform": self.platform,
        }


def get_one(
    code_repository_id: uuid.UUID,
    project_id: uuid.UUID,
    token: str,
) -> CodeRepository:
    """Retrieves a code repository entry from the Mantik API"""

    endpoint = f"/projects/{str(project_id)}/code/{str(code_repository_id)}"
    response = client.send_request_to_mantik_api(
        method="GET", data={}, url_endpoint=endpoint, token=token
    )
    return CodeRepository.from_json(response.json())



import functools

import firecrest
import pytest

import mantik.testing as testing
import mantik_compute_backend.firecrest._exceptions as _exceptions
import mantik_compute_backend.firecrest.connect as _connect


@pytest.mark.parametrize(
    ("login_successful", "expected"),
    [
        (False, _exceptions.AuthenticationFailedException()),
        (True, testing.firecrest.FakeClient),
    ],
)
def test_create_firecrest_api_connection(
    expect_raise_if_exception, monkeypatch, login_successful, expected
):
    monkeypatch.setattr(
        firecrest,
        "Firecrest",
        testing.firecrest.FakeClient,
    )
    monkeypatch.setattr(
        firecrest,
        "ClientCredentialsAuth",
        functools.partial(
            testing.firecrest.FakeClientCredentialsAuth,
            login_successful=login_successful,
        ),
    )
    with expect_raise_if_exception(expected):
        result = _connect.create_firecrest_api_connection(
            token_url="test-token-url",
            api_url="test-api-url",
            user="test-user",
            password="test-password",
        )
        assert isinstance(result, expected)



import re
import urllib.parse

import requests


def remove_double_slashes_from_path(url: str, ensure_https: bool = True) -> str:
    """Ensure that a URL uses HTTPS and does not contain `//` in its path."""
    u = urllib.parse.urlparse(url)
    path = u.path.replace("//", "/")
    url = f"{u.netloc}{path}".replace("//", "/")
    if ensure_https:
        return f"https://{url}"
    return f"{u.scheme}://{url}"


def replace_first_subdomain(url: str, replace_with: str) -> str:
    regex = re.compile(r"^https?:\/\/(www\.)?(.*?)(\..*)$")
    return regex.sub(rf"https://\1{replace_with}\3", url)


def download_from_url(url: str, target_path: str):
    """
    Download artifacts stored at a certain url and saves them in target path.
    """
    file = requests.get(url, stream=True)

    with open(target_path, "wb") as f:
        for chunk in file:
            f.write(chunk)


def get_local_path_from_url(url: str, target_dir: str, filetype: str) -> str:
    filename = url.split("/")[-1].split(filetype)[0] + filetype
    if not target_dir[-1] == "/":
        target_dir += "/"
    return target_dir + filename



import dataclasses
import typing as t
import uuid

import mantik.utils.mantik_api.client as client


@dataclasses.dataclass
class DataRepository:
    uri: str
    is_dvc_enabled: bool
    dvc_connection_id: t.Optional[uuid.UUID]
    connection_id: t.Optional[uuid.UUID]
    versions: t.Dict[str, str]
    platform: str

    @classmethod
    def from_json(cls, response: dict):
        return cls(
            uri=response["uri"],
            is_dvc_enabled=response["isDvcEnabled"],
            dvc_connection_id=uuid.UUID(response["dvcConnectionId"])
            if response.get("dvcConnectionId")
            else None,
            connection_id=uuid.UUID(response["connectionId"])
            if response.get("connectionId")
            else None,
            versions=response["versions"] or {},
            platform=response["platform"],
        )


def get_all(
    project_id: uuid.UUID,
    token: str,
) -> t.List[t.Dict]:
    endpoint = f"/projects/{str(project_id)}/data"
    response = client.send_request_to_mantik_api(
        method="GET", data={}, url_endpoint=endpoint, token=token
    )
    return response.json()["dataRepositories"]


def get_one(
    project_id: uuid.UUID,
    data_repository_id: uuid.UUID,
    token: str,
) -> DataRepository:
    endpoint = f"/projects/{str(project_id)}/data/{str(data_repository_id)}"
    response = client.send_request_to_mantik_api(
        method="GET", data={}, url_endpoint=endpoint, token=token
    )
    return DataRepository.from_json(response.json())



import copy
import dataclasses
import os
import pathlib
import re
import typing as t

import mantik.config._base as _base
import mantik.config._utils as _utils
import mantik.config.exceptions as exceptions
import mantik.config.executable as _executable
import mantik.config.resources as _resources
import mantik.utils as utils
import mantik.utils.mlflow as mlflow

_MLFLOW_ENV_VAR_PREFIX = "MLFLOW_"
_ALLOWED_EXECUTABLES = {
    "Apptainer": _executable.Apptainer,
    "Python": _executable.Python,
}


@dataclasses.dataclass
class Environment(_base.ConfigObject):
    """Part of the backend-config where all variables
    concerning the running environment are stored."""

    execution: t.Optional[_executable.Execution] = None
    variables: t.Optional[t.Dict] = None
    modules: t.Optional[t.List] = None
    pre_run_command_on_login_node: t.Optional[list] = None
    pre_run_command_on_compute_node: t.Optional[list] = None
    post_run_command_on_compute_node: t.Optional[list] = None
    post_run_command_on_login_node: t.Optional[list] = None

    @classmethod
    def _from_dict(cls, config: t.Dict) -> "Environment":
        execution = _get_execution_environment(config)
        variables = _utils.get_optional_config_value(
            name="Variables",
            value_type=dict,
            config=config,
        )
        modules = _utils.get_optional_config_value(
            name="Modules",
            value_type=list,
            config=config,
        )

        preRunCommandOnLoginNode = _utils.get_optional_config_value(
            name="PreRunCommandOnLoginNode",
            value_type=list,
            config=_cast_run_command_str_to_list(
                config, "PreRunCommandOnLoginNode"
            ),
        )
        preRunCommandOnComputeNode = _utils.get_optional_config_value(
            name="PreRunCommandOnComputeNode",
            value_type=list,
            config=_cast_run_command_str_to_list(
                config, "PreRunCommandOnComputeNode"
            ),
        )
        postRunCommandOnComputeNode = _utils.get_optional_config_value(
            name="PostRunCommandOnComputeNode",
            value_type=list,
            config=_cast_run_command_str_to_list(
                config, "PostRunCommandOnComputeNode"
            ),
        )
        postRunCommandOnLoginNode = _utils.get_optional_config_value(
            name="PostRunCommandOnLoginNode",
            value_type=list,
            config=_cast_run_command_str_to_list(
                config, "PostRunCommandOnLoginNode"
            ),
        )

        return cls(
            execution=execution,
            modules=modules,
            variables=variables,
            pre_run_command_on_login_node=preRunCommandOnLoginNode,
            pre_run_command_on_compute_node=preRunCommandOnComputeNode,
            post_run_command_on_compute_node=postRunCommandOnComputeNode,
            post_run_command_on_login_node=postRunCommandOnLoginNode,
        )

    def __post_init__(self):
        """Add all MLflow environment variables to the environment."""
        self.variables = _add_mlflow_env_vars(self.variables)

    def _to_dict(self) -> t.Dict:
        return {
            "Environment": self.variables,
            "User precommand": self._create_login_node_command(
                self.pre_run_command_on_login_node
            ),
            "RunUserPrecommandOnLoginNode": True,
            "Arguments": [],
            "User postcommand": self._create_login_node_command(
                self.post_run_command_on_login_node
            ),
            "RunUserPostcommandOnLoginNode": True,
        }

    def _create_login_node_command(
        self, commands: t.Optional[list]
    ) -> t.Optional[str]:
        # Login node commands are only allowed by UNICORE, and it allows
        # only single strings as commands. Thus, we concatenate multiple
        # commands using `&&`. This makes the command fail if any of
        # the given commands fails.
        return _concatenate_commands(commands, concatenation_str=" && ")

    def _create_compute_node_command(
        self, commands: t.Optional[list]
    ) -> t.Optional[str]:
        return _concatenate_commands(commands, concatenation_str="\n")

    def _create_pre_run_command_on_compute_node(self) -> t.Optional[str]:
        # Venv MUST be loaded before modules
        # see https://gitlab.com/mantik-ai/mantik/issues/140
        return (
            self._create_compute_node_command(
                [
                    self._create_modules_command(),
                    *self.pre_run_command_on_compute_node,
                ],
            )
            if self.pre_run_command_on_compute_node
            else self._create_modules_command()
        )

    def _create_modules_command(self) -> t.Optional[str]:
        return f"module load {' '.join(self.modules)}" if self.modules else None

    def _create_execution_command(self) -> t.Optional[str]:
        if self.execution is not None:
            execution_command = self.execution.get_execution_command()
            arguments = self.execution.get_arguments()
            joined_str = " ".join(filter(None, [execution_command, arguments]))
            return joined_str or None

    def set_srun_cpus_per_task_if_unset(
        self, resources: _resources.Resources
    ) -> "Environment":
        cpus_per_task = resources.cpus_per_node

        if cpus_per_task is not None:
            if self.variables is None:
                self.variables = {"SRUN_CPUS_PER_TASK": str(cpus_per_task)}
            elif (
                self.variables is not None
                and "SRUN_CPUS_PER_TASK" not in self.variables
            ):
                self.variables["SRUN_CPUS_PER_TASK"] = str(cpus_per_task)

        return self

    def _create_execution_command_with_arguments(
        self, entry_point_arguments
    ) -> str:
        if isinstance(self.execution, _executable.Apptainer):
            # For Apptainer, the `apptainer run` command needs
            # to be joined with the MLproject entry point command
            # with a space to get e.g.
            # `apptainer run image.sif python main.py ...`
            concatenation_string = " "
        else:
            # For Python, the `source <path to venv>/bin/activate`
            # command needs to be joined with the MLproject entry
            # point command using `&&` to get
            # `source /path/venv/bin/activate && python main.py ...`
            # (Using `;` will let the job succeed even if one of the
            # commands fails.)
            concatenation_string = " && "
        execution_command_arguments_string = concatenation_string.join(
            filter(
                None, [self._create_execution_command(), entry_point_arguments]
            )
        )
        list_to_join = [
            self._create_pre_run_command_on_compute_node(),
            execution_command_arguments_string,
            self._create_compute_node_command(
                self.post_run_command_on_compute_node
            ),
        ]
        return " && ".join(filter(None, list_to_join))

    def to_unicore_job_description(self, bash_script_name: str) -> t.Dict:
        """Convert to UNICORE job description.

        Parameters
        ----------
        arguments : str
            Arguments to pass to the executable.

        Returns
        -------
        dict
            The UNICORE job description.

            For details see https://sourceforge.net/p/unicore/wiki/Job_Description  # noqa: E501


        """
        job_description = {
            **self.to_dict(),
            "Executable": f'source $(dirname "$(realpath "$0")")/{bash_script_name}',  # noqa E501
        }
        job_description["Environment"] = job_description.get("Environment", {})
        job_description["Environment"].update(
            {"MANTIK_WORKING_DIRECTORY": "$UC_WORKING_DIRECTORY"}
        )

        return job_description

    def execution_given(self) -> bool:
        return self.execution is not None

    def add_env_vars(self, env_vars: t.Dict) -> None:
        if self.variables is None:
            self.variables = env_vars
        else:
            self.variables.update(env_vars)

    def to_slurm_batch_script(
        self, entry_point_arguments: str, run_dir: pathlib.Path
    ) -> str:
        # Setting env vars via the firecREST API is only supported from
        # version >= v1.14.0. Hence, we set them explicitly.
        env_vars_list = [f"export MANTIK_WORKING_DIRECTORY={run_dir}"]
        if self.variables is not None:
            env_vars_list.extend(
                [
                    f"export {key}={value}"
                    for key, value in self.variables.items()
                ]
            )
        env_vars_str = _concatenate_commands(env_vars_list, "\n")

        job_script = _concatenate_commands(
            commands=[
                env_vars_str,
                self._create_execution_command_with_arguments(
                    entry_point_arguments
                ),
            ],
            concatenation_str="\n",
        )
        return job_script

    def to_bash_script(self, entry_point_arguments: str) -> str:
        """Create a bash script for UNICORE."""
        job_script = _concatenate_commands(
            commands=[
                "echo 'Submitted bash script'",
                'cat "$(realpath "$0")"',
                self._create_execution_command_with_arguments(
                    entry_point_arguments
                ),
            ],
            concatenation_str="\n",
        )
        return job_script


def _get_execution_environment(
    config: t.Dict,
) -> t.Optional[_executable.Execution]:
    envs = [env for env in _ALLOWED_EXECUTABLES if env in config]
    execution = _get_only_one_environment_or_none(envs)
    if execution is not None:
        return execution.from_dict(config)
    return


def _get_only_one_environment_or_none(
    env_found: t.List,
) -> t.Optional[t.Type[_executable.Execution]]:
    if not env_found:
        return None
    elif len(env_found) > 1:
        raise exceptions.ConfigValidationError(
            "Only one execution environment is allowed, "
            "but in config these have been found: "
            f"{utils.formatting.iterable_to_string(env_found)}."
        )

    try:
        return _ALLOWED_EXECUTABLES[env_found[0]]
    except KeyError as e:
        raise ValueError(
            f"Environment of type {env_found} not supported"
        ) from e


def _add_mlflow_env_vars(environment: t.Optional[t.Dict]) -> t.Optional[t.Dict]:
    mlflow_env_vars = _get_mlflow_env_vars()
    if mlflow_env_vars:
        if environment is None:
            return mlflow_env_vars
        return {**mlflow_env_vars, **environment}
    return environment


def _get_mlflow_env_vars() -> t.Dict:
    pattern = re.compile(rf"{_MLFLOW_ENV_VAR_PREFIX}\w+")
    return {
        key: value
        for key, value in os.environ.items()
        if pattern.match(key) and key not in mlflow.CONFLICTING_ENV_VARS
    }


def _concatenate_commands(
    commands: t.Optional[list], concatenation_str: str
) -> t.Optional[str]:
    return concatenation_str.join(filter(None, commands)) if commands else None


def _cast_run_command_str_to_list(
    config: t.Dict, run_command_name: str
) -> t.Dict:
    """
    Converts the Pre/PostRunCommands specified in the config from a string
    to list format. This ensures the backward compatibility with configs
    that allow strings in the run commands.
    """
    config = copy.deepcopy(config)
    run_command_value = config.get(run_command_name, None)
    if run_command_value is not None:
        if isinstance(run_command_value, str):
            config[run_command_name] = [run_command_value]
            return config
        elif isinstance(run_command_value, list):
            config[run_command_name] = [
                str(command) for command in run_command_value
            ]
            return config
        else:
            raise exceptions.ConfigValidationError(
                f"Config value for {run_command_name!r} has to be of type str "
                "or list"
            )
    return config



import typing as t
import uuid

import click

import mantik.cli._options as _main_options
import mantik.cli.main as main
import mantik.cli.remote_file_service._arguments as _arguments
import mantik.remote_file_service.data_client as data_client

GROUP_NAME = "s3-file-service"


@main.cli.group(GROUP_NAME)
def cli() -> None:
    """Interaction with the s3 file service."""


@cli.command("copy-file")
@_arguments.SOURCE
@_arguments.TARGET
@_main_options.get_connection_id(required=False)
def copy_file(
    source: str,
    target: str,
    connection_id: t.Optional[uuid.UUID],
) -> None:
    s3_file_service = _import_s3_file_service()
    s3_fs = data_client.DataClient.from_env(
        connection_id=connection_id,
        remote_fs_type=s3_file_service.S3FileService,
    )
    file = s3_fs.copy_file(
        source=source,
        target=target,
    )
    click.echo(file)


@cli.command("remove-file")
@_arguments.TARGET
@_main_options.get_connection_id(required=False)
def remove_file(
    target: str,
    connection_id: t.Optional[uuid.UUID],
) -> None:
    s3_file_service = _import_s3_file_service()
    s3_fs = data_client.DataClient.from_env(
        connection_id=connection_id,
        remote_fs_type=s3_file_service.S3FileService,
    )
    s3_fs.remove_file(target=target)
    click.echo("File removed!")


@cli.command("create-file")
@_arguments.TARGET
@_main_options.get_connection_id(required=False)
def create_file_if_not_exists_file(
    target: str,
    connection_id: t.Optional[uuid.UUID],
) -> None:
    s3_file_service = _import_s3_file_service()
    s3_fs = data_client.DataClient.from_env(
        connection_id=connection_id,
        remote_fs_type=s3_file_service.S3FileService,
    )
    file = s3_fs.create_file_if_not_exists(target=target)
    click.echo(file)


@cli.command("list-directory")
@_arguments.TARGET
@_main_options.get_connection_id(required=False)
def list_directory(target: str, connection_id: t.Optional[uuid.UUID]) -> None:
    s3_file_service = _import_s3_file_service()
    s3_fs = data_client.DataClient.from_env(
        connection_id=connection_id,
        remote_fs_type=s3_file_service.S3FileService,
    )
    directory = s3_fs.list_directory(target=target)
    click.echo(directory)


@cli.command("create-directory")
@_arguments.TARGET
@_main_options.get_connection_id(required=False)
def create_directory(
    target: str,
    connection_id: t.Optional[uuid.UUID],
) -> None:
    s3_file_service = _import_s3_file_service()
    s3_fs = data_client.DataClient.from_env(
        connection_id=connection_id,
        remote_fs_type=s3_file_service.S3FileService,
    )
    directory = s3_fs.create_directory(target=target)
    click.echo(directory)


@cli.command("remove-directory")
@_arguments.TARGET
@_main_options.get_connection_id(required=False)
def remove_directory(
    target: str,
    connection_id: t.Optional[uuid.UUID],
) -> None:
    s3_file_service = _import_s3_file_service()
    s3_fs = data_client.DataClient.from_env(
        connection_id=connection_id,
        remote_fs_type=s3_file_service.S3FileService,
    )
    s3_fs.remove_directory(target=target)
    click.echo("Directory removed!")


@cli.command("copy-directory")
@_arguments.SOURCE
@_arguments.TARGET
@_main_options.get_connection_id(required=False)
def copy_directory(
    source: str,
    target: str,
    connection_id: t.Optional[uuid.UUID],
) -> None:
    s3_file_service = _import_s3_file_service()
    s3_fs = data_client.DataClient.from_env(
        connection_id=connection_id,
        remote_fs_type=s3_file_service.S3FileService,
    )
    directory = s3_fs.copy_directory(
        source=source,
        target=target,
    )
    click.echo(directory)


@cli.command("exists")
@_arguments.TARGET
@_main_options.get_connection_id(required=False)
def exists(target: str, connection_id: t.Optional[uuid.UUID]) -> None:
    s3_file_service = _import_s3_file_service()
    s3_fs = data_client.DataClient.from_env(
        connection_id=connection_id,
        remote_fs_type=s3_file_service.S3FileService,
    )
    _exists = s3_fs.exists(target=target)
    click.echo(_exists)


@cli.command("user")
@_main_options.get_connection_id(required=False)
def user(connection_id: t.Optional[uuid.UUID]) -> None:
    s3_file_service = _import_s3_file_service()
    s3_fs = data_client.DataClient.from_env(
        connection_id=connection_id,
        remote_fs_type=s3_file_service.S3FileService,
    )
    click.echo(s3_fs.user)


def _import_s3_file_service():
    try:
        import mantik.remote_file_service.s3_file_service as s3_file_service
    except ModuleNotFoundError as e:
        raise RuntimeError(
            "Extras for S3 file service must be "
            'installed via `pip install "mantik[s3]"`'
        ) from e
    else:
        return s3_file_service



import dataclasses
import typing as t
import uuid

import jose.jwt

import mantik.authentication as authentication
import mantik.utils.env as env
import mantik.utils.mantik_api.connection

UNICORE_USERNAME_ENV_VAR = "MANTIK_UNICORE_USERNAME"
UNICORE_PASSWORD_ENV_VAR = "MANTIK_UNICORE_PASSWORD"
FIRECREST_CLIENT_ID_ENV_VAR = "MANTIK_FIRECREST_CLIENT_ID"
FIRECREST_CLIENT_SECRET_ENV_VAR = "MANTIK_FIRECREST_CLIENT_SECRET"


@dataclasses.dataclass
class HpcRestApi:
    username: str
    password: str

    @classmethod
    def from_unicore_env_vars(
        cls,
        connection_id: t.Optional[uuid.UUID] = None,
    ) -> "HpcRestApi":
        if connection_id:
            return cls._credentials_from_api(connection_id=connection_id)
        return cls._credentials_from_env_vars(
            username_env_var=UNICORE_USERNAME_ENV_VAR,
            password_env_var=UNICORE_PASSWORD_ENV_VAR,
        )

    @classmethod
    def from_firecrest_env_vars(
        cls,
        connection_id: t.Optional[uuid.UUID] = None,
    ) -> "HpcRestApi":
        if connection_id:
            return cls._credentials_from_api(connection_id=connection_id)
        return cls._credentials_from_env_vars(
            username_env_var=FIRECREST_CLIENT_ID_ENV_VAR,
            password_env_var=FIRECREST_CLIENT_SECRET_ENV_VAR,
        )

    @classmethod
    def _credentials_from_api(
        cls,
        connection_id: uuid.UUID,
    ) -> "HpcRestApi":
        access_token = authentication.auth.get_valid_access_token()
        user_id = _get_sub_from_token(access_token)
        connection = mantik.utils.mantik_api.connection.get(
            user_id=uuid.UUID(user_id),
            connection_id=connection_id,
            token=access_token,
        )
        return cls(username=connection.login_name, password=connection.password)

    @classmethod
    def _credentials_from_env_vars(
        cls, username_env_var: str, password_env_var: str
    ) -> "HpcRestApi":
        username = env.get_required_env_var(username_env_var)
        password = env.get_required_env_var(password_env_var)
        return cls(username=username, password=password)


def _get_sub_from_token(token: str):
    # The `sub` field of the token claims contains the user UUID.
    return jose.jwt.get_unverified_claims(token)["sub"]



import logging
import pathlib

import mantik.authentication.api as api
import mantik.authentication.tokens as _tokens

logger = logging.getLogger(__name__)

_MANTIK_FOLDER = pathlib.Path.home() / ".mantik"
_MANTIK_TOKEN_FILE = _MANTIK_FOLDER / "tokens.json"


def get_valid_access_token() -> str:
    """Authenticate at Mantik.

    Returns
    -------
    Access Token: str

    Notes
    -----
    Mantik prioritizes the username and password environment variables over
    the token variable, causing an `Unauthorized` error. As a consequence,
    these have to be unset before setting the token variable.

    The tokens will be stored in a file `~/.mantik/tokens.json` to reuse
    tokens and refresh them only if they have expired.

    """
    logger.debug("Getting tokens")
    tokens = _get_and_store_tokens()
    return tokens.access_token


def _get_and_store_tokens() -> _tokens.Tokens:
    if not _MANTIK_TOKEN_FILE.exists():
        logger.debug(
            "No existing tokens file found at %s", _MANTIK_TOKEN_FILE.as_posix()
        )
        return _create_and_store_new_tokens()
    return _read_stored_tokens()


def _create_and_store_new_tokens() -> _tokens.Tokens:
    logger.debug(
        "Creating new tokens to store to %s", _MANTIK_TOKEN_FILE.as_posix()
    )
    tokens = api.create_tokens()
    tokens.write_to_file(_MANTIK_TOKEN_FILE)
    return tokens


def _read_stored_tokens() -> _tokens.Tokens:
    logger.debug(
        "Reading stored tokens from file %s", _MANTIK_TOKEN_FILE.as_posix()
    )
    tokens = _tokens.Tokens.from_file(_MANTIK_TOKEN_FILE)
    if tokens.has_expired:
        logger.debug("Access token has expired")
        refreshed = api.refresh_tokens(tokens=tokens)
        refreshed.write_to_file(_MANTIK_TOKEN_FILE)
        return refreshed
    return tokens



import logging

import mantik.authentication.auth as auth
import mantik.tracking.environment as _environment
import mantik.utils as utils

logger = logging.getLogger(__name__)


def init_tracking() -> _environment.Environment:
    """Authenticate to the MLflow Tracking Server.

    Returns
    -------
    Environment
        Holds the environment variables required for tracking.

    Notes
    -----
    MLflow prioritizes the username and password environment variables over
    the token variable, causing an `Unauthorized` error. As a consequence,
    these have to be unset before setting the token variable.

    The tokens will be stored in a file `~/.mantik/tokens.json` to reuse
    tokens and refresh them only if they have expired.

    """
    logger.debug("Authenticating at the Mantik platform")
    token = auth.get_valid_access_token()
    environment = _environment.Environment(token=token)
    utils.mlflow.unset_conflicting_env_vars()
    utils.env.set_env_vars(environment.to_dict())
    logger.debug("Authentication successful")
    return environment



import logging
import uuid

import mantik.cli.utils as utils
import mantik.utils.mantik_api.models
import mantik.utils.urls

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)


def download_container(
    project_id: uuid.UUID,
    model_id: uuid.UUID,
    target_dir: str,
    image_type: str,
    load: bool,
):
    if image_type.lower() == "docker":
        image_url = mantik.utils.mantik_api.models.get_image_url(
            project_id=project_id,
            model_id=model_id,
            token=utils.access_token_from_env_vars(),
        )

        image_path = mantik.utils.urls.get_local_path_from_url(
            target_dir=target_dir, url=image_url, filetype=".tar.gz"
        )
        _save_image(image_path=image_path, image_url=image_url)

        if load:
            _load_to_docker(image_path=image_path)
        else:
            logger.info(
                "In order to load the image into docker, "
                f"you can run `docker load < {image_path}`"
            )
    else:
        logger.error(f"We are sorry, {image_type} is not supported yet!")
        raise NotImplementedError


def _save_image(image_path: str, image_url: str):
    logger.info(
        f"Downloading image: {image_path}. This can take a few minutes..."
    )
    mantik.utils.urls.download_from_url(url=image_url, target_path=image_path)
    logger.info(f"Image saved as zipped tarball at {image_path}")


def _load_to_docker(image_path: str):
    try:
        import docker
    except ModuleNotFoundError as exc:
        raise RuntimeError(
            "Install supported 'docker' version "
            "(`pip install mantik[docker]`)"
            "or any desired 'docker' python package version"
        ) from exc
    logger.info(
        "Unzipping and loading into docker. This can take a few minutes..."
    )

    client = docker.from_env()

    with open(image_path, "rb") as f:
        client.images.load(f)

    filename = image_path.split("/")[-1][:-7]
    logger.info(
        "Run 'docker images', an image named " f"{filename} should be present."
    )

    logger.info(
        f"Run 'docker run -p 8080:8080 {filename}` "
        "to serve the model image for inference."
    )



import datetime
import functools
import typing as t

import mantik.authentication.auth as auth
import mantik.authentication.tokens as _tokens


def set_token(
    access_token: str = "test-access-token",
    refresh_token: str = "test-refresh-token",
    expires_at: datetime.datetime = datetime.datetime(2020, 1, 1, 2),
) -> t.Callable:
    import freezegun

    def decorate(func):
        @functools.wraps(func)
        @freezegun.freeze_time("2020-01-01T01:00")
        def wrapper(*args, **kwargs):
            # The client calls `mantik.tracking.init_tracking()`.
            # Thus, we store valid, non-expired tokens before to
            # avoid a refreshing.
            tokens = _tokens.Tokens(
                access_token=access_token,
                refresh_token=refresh_token,
                expires_at=expires_at,
            )
            tokens.write_to_file(auth._MANTIK_TOKEN_FILE)
            return func(*args, **kwargs)

        return wrapper

    return decorate



import pathlib

import pytest

import mantik.config
import mantik.config.exceptions as exceptions
import mantik.testing as testing


class TestConfig:
    @pytest.mark.parametrize(
        ("env_vars", "d", "expected", "expected_messgage"),
        [
            # Test case: Neither UnicoreApiUrl nor Firecrest section given
            (
                testing.config.ALL_UNICORE_ENV_VARS,
                {},
                exceptions.ConfigValidationError(),
                "Either UnicoreApiUrl or Firecrest details must be provided",
            ),
            # Test cases: UNICORE environment variables not set.
            (
                [],
                {
                    "UnicoreApiUrl": "test-url",
                    "Resources": {"Queue": "batch"},
                },
                NameError(),
                "Environment variable 'MANTIK_UNICORE_USERNAME' not set",
            ),
            (
                testing.config.ALL_UNICORE_ENV_VARS[:1],
                {
                    "UnicoreApiUrl": "test-url",
                    "Resources": {"Queue": "batch"},
                },
                NameError(),
                "Environment variable 'MANTIK_UNICORE_PASSWORD' not set",
            ),
            # Test cases: firecREST environment variables not set.
            (
                [],
                {
                    "Firecrest": {
                        "ApiUrl": "test-api-url",
                        "TokenUrl": "test-token-url",
                        "Machine": "test-machine",
                    },
                    "Resources": {"Queue": "batch"},
                },
                NameError(),
                "Environment variable 'MANTIK_FIRECREST_CLIENT_ID' not set",
            ),
            (
                testing.config.ALL_FIRECREST_ENV_VARS[:1],
                {
                    "Firecrest": {
                        "ApiUrl": "test-api-url",
                        "TokenUrl": "test-token-url",
                        "Machine": "test-machine",
                    },
                    "Resources": {"Queue": "batch"},
                },
                NameError(),
                "Environment variable 'MANTIK_FIRECREST_CLIENT_SECRET' not set",
            ),
            # Test cases: No compute budget account env var set
            (
                testing.config.ALL_UNICORE_ENV_VARS[:2],
                {
                    "UnicoreApiUrl": "test-url",
                    "Resources": {"Queue": "batch"},
                },
                NameError(),
                "Environment variable 'MANTIK_COMPUTE_BUDGET_ACCOUNT' not set",
            ),
            (
                testing.config.ALL_FIRECREST_ENV_VARS[:2],
                {
                    "Firecrest": {
                        "ApiUrl": "test-api-url",
                        "TokenUrl": "test-token-url",
                        "Machine": "test-machine",
                    },
                    "Resources": {"Queue": "batch"},
                },
                NameError(),
                "Environment variable 'MANTIK_COMPUTE_BUDGET_ACCOUNT' not set",
            ),
            # Test case: backend config missing Resources section.
            (
                testing.config.ALL_UNICORE_ENV_VARS,
                {"Queue": "batch"},
                exceptions.ConfigValidationError(),
                "Either UnicoreApiUrl or Firecrest details must be provided",
            ),
            # Test case: Neither unicore nor firecrest given.
            (
                testing.config.ALL_UNICORE_ENV_VARS,
                {
                    "Environment": {
                        "Apptainer": {
                            "Path": "image.sif",
                        },
                    },
                    "Resources": {"Queue": "batch"},
                },
                exceptions.ConfigValidationError(),
                "Either UnicoreApiUrl or Firecrest details must be provided",
            ),
            # Test case: Both unicore and firecrest given
            (
                testing.config.ALL_FIRECREST_ENV_VARS,
                {
                    "UnicoreApiUrl": "test-url",
                    "Firecrest": {
                        "ApiUrl": "test-api-url",
                        "TokenUrl": "test-token-url",
                        "Machine": "test-machine",
                    },
                    "Environment": {
                        "Apptainer": {
                            "Path": "image.sif",
                        },
                    },
                    "Resources": {"Queue": "batch"},
                },
                exceptions.ConfigValidationError(),
                (
                    "Either UnicoreApiUrl or Firecrest details must be "
                    "provided, but both are given"
                ),
            ),
            # Test case: Remote Apptainer image given but relative path.
            (
                testing.config.ALL_UNICORE_ENV_VARS,
                {
                    "UnicoreApiUrl": "test-url",
                    "Resources": {"Queue": "batch"},
                    "Environment": {
                        "Apptainer": {
                            "Path": "../relative/path/to/image.sif",
                            "Type": "remote",
                        },
                    },
                },
                mantik.config.exceptions.ConfigValidationError(),
                (
                    "If image type 'remote' is given for an Apptainer image, "
                    "the given path must be absolute. The given path is: "
                    "'../relative/path/to/image.sif'."
                ),
            ),
            # Test case: Config entry has incorrect type.
            (
                testing.config.ALL_UNICORE_ENV_VARS,
                {
                    "UnicoreApiUrl": "test-url",
                    "Resources": {"Queue": "batch", "Nodes": "incorrect type"},
                },
                mantik.config.exceptions.ConfigValidationError(),
                "Config value for 'Nodes' has to be of type int",
            ),
        ],
    )
    @pytest.mark.usefixtures("unset_tracking_token_env_var_before_execution")
    def test_from_dict_raises(
        self,
        expect_raise_if_exception,
        env_vars_set,
        env_vars,
        d,
        expected,
        expected_messgage,
    ):
        env_vars = {
            var: testing.config.DEFAULT_ENV_VAR_VALUE for var in env_vars
        }

        with env_vars_set(env_vars), expect_raise_if_exception(expected) as e:
            mantik.config.core.Config.from_dict(d)

        result = str(e.value)

        assert result == expected_messgage, "Incorrect error message raised"

    @pytest.mark.parametrize(
        ("env_vars", "d", "expected"),
        [
            # Test case: Everything correct.
            (
                testing.config.ALL_UNICORE_ENV_VARS,
                {
                    "UnicoreApiUrl": "test-url",
                    "Environment": {
                        "Apptainer": {
                            "Path": "image.sif",
                        },
                    },
                    "Resources": {"Queue": "batch"},
                },
                testing.config._create_config_with_unicore(
                    resources=mantik.config.resources.Resources(queue="batch"),
                    env=mantik.config.environment.Environment(
                        execution=mantik.config.executable.Apptainer(
                            path=pathlib.Path("image.sif"),
                        ),
                    ),
                ),
            ),
            # Test case: Using firecrest
            (
                testing.config.ALL_FIRECREST_ENV_VARS,
                {
                    "Firecrest": {
                        "ApiUrl": "test-api-url",
                        "TokenUrl": "test-token-url",
                        "Machine": "test-machine",
                    },
                    "Environment": {
                        "Apptainer": {
                            "Path": "image.sif",
                        },
                    },
                    "Resources": {"Queue": "batch", "NodeConstraints": "gpu"},
                },
                testing.config._create_config_with_firecrest(
                    resources=mantik.config.resources.Resources(
                        queue="batch", node_constraints="gpu"
                    ),
                    env=mantik.config.environment.Environment(
                        execution=mantik.config.executable.Apptainer(
                            path=pathlib.Path("image.sif"),
                        ),
                    ),
                ),
            ),
            # Test case: Using firecrest
            (
                testing.config.ALL_FIRECREST_ENV_VARS,
                {
                    "Firecrest": {
                        "ApiUrl": "test-api-url",
                        "TokenUrl": "test-token-url",
                        "Machine": "test-machine",
                    },
                    "Environment": {
                        "Apptainer": {
                            "Path": "image.sif",
                        },
                    },
                    "Resources": {"Queue": "batch", "NodeConstraints": "gpu"},
                },
                testing.config._create_config_with_firecrest(
                    resources=mantik.config.resources.Resources(
                        queue="batch", node_constraints="gpu"
                    ),
                    env=mantik.config.environment.Environment(
                        execution=mantik.config.executable.Apptainer(
                            path=pathlib.Path("image.sif"),
                        ),
                    ),
                ),
            ),
            # Test case: No environment given
            (
                testing.config.ALL_UNICORE_ENV_VARS,
                {
                    "UnicoreApiUrl": "test-url",
                    "Resources": {"Queue": "batch"},
                },
                testing.config._create_config_with_unicore(
                    resources=mantik.config.resources.Resources(queue="batch"),
                ),
            ),
            # Test case: No environment given, but resources
            (
                testing.config.ALL_UNICORE_ENV_VARS,
                {
                    "UnicoreApiUrl": "test-url",
                    "Resources": {
                        "Queue": "batch",
                        "Runtime": "2h",
                        "Nodes": 1,
                        "TotalCPUs": 2,
                        "CPUsPerNode": 3,
                        "GPUsPerNode": 4,
                        "MemoryPerNode": "10G",
                        "Reservation": "test-reservation",
                        "NodeConstraints": "mem192",
                        "QoS": "test-qos",
                    },
                },
                testing.config._create_config_with_unicore(
                    resources=mantik.config.resources.Resources(
                        queue="batch",
                        runtime="2h",
                        nodes=1,
                        total_cpus=2,
                        cpus_per_node=3,
                        gpus_per_node=4,
                        memory_per_node="10G",
                        reservation="test-reservation",
                        node_constraints="mem192",
                        qos="test-qos",
                    ),
                ),
            ),
            # Test case: Only modules given in environment
            (
                testing.config.ALL_UNICORE_ENV_VARS,
                {
                    "UnicoreApiUrl": "test-url",
                    "Environment": {
                        "Modules": ["module"],
                    },
                    "Resources": {"Queue": "batch"},
                },
                testing.config._create_config_with_unicore(
                    resources=mantik.config.resources.Resources(queue="batch"),
                    env=mantik.config.environment.Environment(
                        modules=["module"],
                    ),
                ),
            ),
            # Test case: Local Apptainer image given.
            (
                testing.config.ALL_UNICORE_ENV_VARS,
                {
                    "UnicoreApiUrl": "test-url",
                    "Resources": {"Queue": "batch"},
                    "Environment": {
                        "Apptainer": {
                            "Path": "image.sif",
                        },
                    },
                },
                testing.config._create_config_with_unicore(
                    resources=mantik.config.resources.Resources(queue="batch"),
                    env=mantik.config.environment.Environment(
                        execution=mantik.config.executable.Apptainer(
                            path=pathlib.Path("image.sif"),
                        ),
                    ),
                ),
            ),
            # Test case: Local Apptainer image given with additional options
            # as list of strings
            (
                testing.config.ALL_UNICORE_ENV_VARS,
                {
                    "UnicoreApiUrl": "test-url",
                    "Resources": {"Queue": "batch"},
                    "Environment": {
                        "Apptainer": {
                            "Path": "image.sif",
                            "Options": ["--nv", "-B /data:/data"],
                        },
                    },
                },
                testing.config._create_config_with_unicore(
                    resources=mantik.config.resources.Resources(queue="batch"),
                    env=mantik.config.environment.Environment(
                        execution=mantik.config.executable.Apptainer(
                            path=pathlib.Path("image.sif"),
                            options=["--nv", "-B /data:/data"],
                        )
                    ),
                ),
            ),
            # Test case: Local Apptainer image given with additional options
            # as string
            (
                testing.config.ALL_UNICORE_ENV_VARS,
                {
                    "UnicoreApiUrl": "test-url",
                    "Resources": {"Queue": "batch"},
                    "Environment": {
                        "Apptainer": {
                            "Path": "image.sif",
                            "Options": "--nv -B /data:/data",
                        },
                    },
                },
                testing.config._create_config_with_unicore(
                    resources=mantik.config.resources.Resources(queue="batch"),
                    env=mantik.config.environment.Environment(
                        execution=mantik.config.executable.Apptainer(
                            path=pathlib.Path("image.sif"),
                            options=["--nv -B /data:/data"],
                        )
                    ),
                ),
            ),
            # Test case: Remote Apptainer image given.
            (
                testing.config.ALL_UNICORE_ENV_VARS,
                {
                    "UnicoreApiUrl": "test-url",
                    "Resources": {"Queue": "batch"},
                    "Environment": {
                        "Apptainer": {
                            "Path": "/absolute/path/to/image.sif",
                            "Type": "remote",
                        },
                    },
                },
                testing.config._create_config_with_unicore(
                    resources=mantik.config.resources.Resources(queue="batch"),
                    env=mantik.config.environment.Environment(
                        execution=mantik.config.executable.Apptainer(
                            path=pathlib.Path("/absolute/path/to/image.sif"),
                            type="remote",
                        ),
                    ),
                ),
            ),
            # Test case: More variables and modules given.
            (
                testing.config.ALL_UNICORE_ENV_VARS,
                {
                    "UnicoreApiUrl": "test-url",
                    "Resources": {"Queue": "batch"},
                    "Environment": {
                        "Apptainer": {
                            "Path": "/absolute/path/to/image.sif",
                            "Type": "remote",
                        },
                        "Variables": {"TEST_ENV_VAR": "value"},
                        "Modules": [
                            "TensorFlow/2.5.0-Python-3.8.5",
                            "Horovod/0.23.0-Python-3.8.5",
                            "PyTorch/1.8.1-Python-3.8.5",
                        ],
                    },
                },
                testing.config._create_config_with_unicore(
                    resources=mantik.config.resources.Resources(queue="batch"),
                    env=mantik.config.environment.Environment(
                        execution=mantik.config.executable.Apptainer(
                            path=pathlib.Path("/absolute/path/to/image.sif"),
                            type="remote",
                        ),
                        modules=[
                            "TensorFlow/2.5.0-Python-3.8.5",
                            "Horovod/0.23.0-Python-3.8.5",
                            "PyTorch/1.8.1-Python-3.8.5",
                        ],
                        variables={"TEST_ENV_VAR": "value"},
                    ),
                ),
            ),
        ],
    )
    @pytest.mark.usefixtures("unset_tracking_token_env_var_before_execution")
    def test_from_dict(
        self, expect_raise_if_exception, env_vars_set, env_vars, d, expected
    ):
        env_vars = {
            var: testing.config.DEFAULT_ENV_VAR_VALUE for var in env_vars
        }

        with env_vars_set(env_vars):
            result = mantik.config.core.Config.from_dict(d)

            assert result == expected

    @pytest.mark.parametrize(
        "d",
        [
            {
                "Firecrest": {
                    "ApiUrl": "test-api-url",
                    "TokenUrl": "test-token-url",
                    "Machine": "test-machine",
                },
                "Environment": {
                    "PreRunCommandOnLoginNode": "echo hello",
                },
                "Resources": {"Queue": "batch", "NodeConstraints": "gpu"},
            },
            {
                "Firecrest": {
                    "ApiUrl": "test-api-url",
                    "TokenUrl": "test-token-url",
                    "Machine": "test-machine",
                },
                "Environment": {
                    "PostRunCommandOnLoginNode": "echo hello",
                },
                "Resources": {"Queue": "batch", "NodeConstraints": "gpu"},
            },
            {
                "Firecrest": {
                    "ApiUrl": "test-api-url",
                    "TokenUrl": "test-token-url",
                    "Machine": "test-machine",
                },
                "Environment": {
                    "PreRunCommandOnLoginNode": "echo hello",
                    "PostRunCommandOnLoginNode": "echo hello",
                },
                "Resources": {"Queue": "batch", "NodeConstraints": "gpu"},
            },
        ],
    )
    def test_command_on_login_node_raises_for_firecrest(
        self, env_vars_set, expect_raise_if_exception, d
    ):
        env_vars = {
            var: testing.config.DEFAULT_ENV_VAR_VALUE
            for var in testing.config.ALL_FIRECREST_ENV_VARS
        }
        expected = "Pre-/PostRunCommandOnLoginNode not supported by firecREST"

        with env_vars_set(env_vars), expect_raise_if_exception(
            exceptions.ConfigValidationError()
        ) as e:
            mantik.config.core.Config.from_dict(d)

        assert str(e.value) == expected

    @pytest.mark.parametrize(
        ("config", "expected"),
        [
            # Test case: Only project, resources and
            # environment are included in the returned dict.
            (
                mantik.config.core.Config(
                    unicore_api_url="not_included",
                    user="not_included",
                    password="not_included",
                    project="test-project",
                    resources=mantik.config.resources.Resources(
                        queue="batch",
                    ),
                    environment=mantik.config.environment.Environment(
                        execution=mantik.config.executable.Apptainer(
                            path=pathlib.Path("test-image")
                        ),
                    ),
                ),
                {
                    "Project": "test-project",
                    "Resources": {"Queue": "batch"},
                    "Arguments": [],
                    "RunUserPrecommandOnLoginNode": True,
                    "RunUserPostcommandOnLoginNode": True,
                    "Stdout": "mantik.log",
                    "Stderr": "mantik.log",
                },
            ),
            # Test case: All given values included.
            (
                mantik.config.core.Config(
                    unicore_api_url="not_included",
                    user="not_included",
                    password="not_included",
                    project="test-project",
                    resources=mantik.config.resources.Resources(
                        queue="batch",
                        runtime="1h",
                        nodes=2,
                        total_cpus=48,
                        cpus_per_node=24,
                        memory_per_node="10G",
                        gpus_per_node=1,
                        reservation="test-reservation",
                        node_constraints="test-node-constraints",
                        qos="test-qos",
                    ),
                    environment=mantik.config.environment.Environment(
                        execution=mantik.config.executable.Apptainer(
                            path=pathlib.Path("test-image"),
                        ),
                        variables={"TEST": "test"},
                    ),
                ),
                {
                    "Project": "test-project",
                    "Resources": {
                        "Queue": "batch",
                        "Runtime": "1h",
                        "Nodes": 2,
                        "TotalCPUs": 48,
                        "CPUsPerNode": 24,
                        "GPUS": 1,
                        "MemoryPerNode": "10G",
                        "Reservation": "test-reservation",
                        "NodeConstraints": "test-node-constraints",
                        "QoS": "test-qos",
                    },
                    "Environment": {"TEST": "test", "SRUN_CPUS_PER_TASK": "24"},
                    "Arguments": [],
                    "RunUserPrecommandOnLoginNode": True,
                    "RunUserPostcommandOnLoginNode": True,
                    "Stdout": "mantik.log",
                    "Stderr": "mantik.log",
                },
            ),
        ],
    )
    def test_to_dict(self, config, expected):
        result = config.to_dict()

        assert result == expected

    @pytest.mark.parametrize(
        ("config", "expected"),
        [
            # Test case: No environment section, and cpus_per_node unset
            (
                mantik.config.core.Config(
                    unicore_api_url="not_included",
                    user="not_included",
                    password="not_included",
                    project="test-project",
                    resources=mantik.config.resources.Resources(
                        queue="batch",
                    ),
                ),
                None,
            ),
            # Test case: Environment section, but no Variables, and
            # cpus_per_node unset
            (
                mantik.config.core.Config(
                    unicore_api_url="not_included",
                    user="not_included",
                    password="not_included",
                    project="test-project",
                    resources=mantik.config.resources.Resources(
                        queue="batch",
                    ),
                    environment=mantik.config.environment.Environment(),  # noqa E501
                ),
                None,
            ),
            # Test case: Environment section with Variables, and
            # cpus_per_node unset
            (
                mantik.config.core.Config(
                    unicore_api_url="not_included",
                    user="not_included",
                    password="not_included",
                    project="test-project",
                    resources=mantik.config.resources.Resources(
                        queue="batch",
                    ),
                    environment=mantik.config.environment.Environment(
                        variables={"TEST_VAR": "test_value"}
                    ),
                ),
                {"TEST_VAR": "test_value"},
            ),
            # Test case: SRUN_CPUS_PER_TASK set explicitly by user,
            # and cpus_per_node unset
            (
                mantik.config.core.Config(
                    unicore_api_url="not_included",
                    user="not_included",
                    password="not_included",
                    project="test-project",
                    resources=mantik.config.resources.Resources(
                        queue="batch",
                    ),
                    environment=mantik.config.environment.Environment(
                        variables={"SRUN_CPUS_PER_TASK": "2"},
                    ),
                ),
                {"SRUN_CPUS_PER_TASK": "2"},
            ),
            # Test case: No environment section, and cpus_per_node set
            (
                mantik.config.core.Config(
                    unicore_api_url="not_included",
                    user="not_included",
                    password="not_included",
                    project="test-project",
                    resources=mantik.config.resources.Resources(
                        queue="batch",
                        cpus_per_node=2,
                    ),
                ),
                {"SRUN_CPUS_PER_TASK": "2"},
            ),
            # Test case: Environment section, but no Variables, and
            # cpus_per_node set
            (
                mantik.config.core.Config(
                    unicore_api_url="not_included",
                    user="not_included",
                    password="not_included",
                    project="test-project",
                    resources=mantik.config.resources.Resources(
                        queue="batch",
                        cpus_per_node=2,
                    ),
                    environment=mantik.config.environment.Environment(),  # noqa E501
                ),
                {"SRUN_CPUS_PER_TASK": "2"},
            ),
            # Test case: Environment section with Variables, and
            # cpus_per_node set
            (
                mantik.config.core.Config(
                    unicore_api_url="not_included",
                    user="not_included",
                    password="not_included",
                    project="test-project",
                    resources=mantik.config.resources.Resources(
                        queue="batch",
                        cpus_per_node=2,
                    ),
                    environment=mantik.config.environment.Environment(
                        variables={"TEST_VAR": "test_value"},
                    ),
                ),
                {"TEST_VAR": "test_value", "SRUN_CPUS_PER_TASK": "2"},
            ),
            # Test case: SRUN_CPUS_PER_TASK set explicitly by user, and
            # cpus_per_node set
            (
                mantik.config.core.Config(
                    unicore_api_url="not_included",
                    user="not_included",
                    password="not_included",
                    project="test-project",
                    resources=mantik.config.resources.Resources(
                        queue="batch",
                        cpus_per_node=1,
                    ),
                    environment=mantik.config.environment.Environment(
                        variables={
                            "TEST_VAR": "test_value",
                            "SRUN_CPUS_PER_TASK": "2",
                        }
                    ),
                ),
                {"TEST_VAR": "test_value", "SRUN_CPUS_PER_TASK": "2"},
            ),
        ],
    )
    def test_with_optional_add_srun_cpus_per_task_to_environment(
        self, config, expected
    ) -> None:
        result = config.environment.variables

        assert result == expected

    @pytest.mark.parametrize(
        ("test_config", "expected"),
        [
            (
                mantik.config.core.Config(
                    unicore_api_url="not_included",
                    user="not_included",
                    password="not_included",
                    project="test-project",
                    resources=mantik.config.resources.Resources(
                        queue="batch",
                        cpus_per_node=1,
                    ),
                    environment=mantik.config.environment.Environment(
                        execution=mantik.config.executable.Apptainer(
                            path=pathlib.Path("test-image"),
                        ),
                        variables={"TEST": "test"},
                        pre_run_command_on_compute_node="echo precommand on compute node",  # noqa E501
                        post_run_command_on_compute_node="echo postcommand on compute node",  # noqa E501
                    ),
                ),
                {
                    "Project": "test-project",
                    "Resources": {"Queue": "batch", "CPUsPerNode": 1},
                    "Stdout": "mantik.log",
                    "Stderr": "mantik.log",
                    "Environment": {
                        "TEST": "test",
                        "SRUN_CPUS_PER_TASK": "1",
                        "MANTIK_WORKING_DIRECTORY": "$UC_WORKING_DIRECTORY",
                    },
                    "RunUserPrecommandOnLoginNode": True,
                    "Executable": 'source $(dirname "$(realpath "$0")")/mantik.sh',  # noqa E501
                    "Arguments": [],
                    "RunUserPostcommandOnLoginNode": True,
                },
            ),
            (
                mantik.config.core.Config(
                    unicore_api_url="not_included",
                    user="not_included",
                    password="not_included",
                    project="test-project",
                    resources=mantik.config.resources.Resources(
                        queue="batch",
                        cpus_per_node=1,
                    ),
                    environment=mantik.config.environment.Environment(
                        execution=mantik.config.executable.Python(
                            path=pathlib.Path("/venv"),
                        ),
                        variables={"TEST": "test"},
                        pre_run_command_on_compute_node="echo precommand on compute node",  # noqa E501
                        post_run_command_on_compute_node="echo postcommand on compute node",  # noqa E501
                    ),
                ),
                {
                    "Project": "test-project",
                    "Resources": {"Queue": "batch", "CPUsPerNode": 1},
                    "Stdout": "mantik.log",
                    "Stderr": "mantik.log",
                    "Environment": {
                        "TEST": "test",
                        "SRUN_CPUS_PER_TASK": "1",
                        "MANTIK_WORKING_DIRECTORY": "$UC_WORKING_DIRECTORY",
                    },
                    "RunUserPrecommandOnLoginNode": True,
                    "Arguments": [],
                    "Executable": 'source $(dirname "$(realpath "$0")")/mantik.sh',  # noqa E501
                    "RunUserPostcommandOnLoginNode": True,
                },
            ),
        ],
    )
    def test_to_job_description(self, test_config, expected):
        bash_script_name = "mantik.sh"
        result = test_config.to_unicore_job_description(bash_script_name)
        assert result == expected



import dataclasses
import enum
import logging
import pathlib
import typing as t
import uuid

import mantik.config._base as _base
import mantik.config._utils as _utils
import mantik.config.environment as _environment
import mantik.config.exceptions as _exceptions
import mantik.config.firecrest as _firecrest
import mantik.config.read as read
import mantik.config.resources as _resources
import mantik.utils.credentials as _credentials
import mantik.utils.env as env

COMPUTE_BUDGET_ACCOUNT_ENV_VAR = "MANTIK_COMPUTE_BUDGET_ACCOUNT"
_UNICORE_AUTH_SERVER_URL_ENV_VAR = "MANTIK_UNICORE_AUTH_SERVER_URL"
APPLICATION_LOGS_FILE = "mantik.log"

logger = logging.getLogger(__name__)


class SupportedBackends(enum.Enum):
    UNICORE = "unicore"
    FIRECREST = "firecrest"


@dataclasses.dataclass
class Config(_base.ConfigObject):
    """The backend config for the UNICORE MLflow backend."""

    user: str
    password: str
    project: str
    resources: _resources.Resources
    environment: t.Optional[_environment.Environment] = None
    exclude: t.Optional[t.List] = None
    unicore_api_url: t.Optional[str] = None
    firecrest: t.Optional[_firecrest.Firecrest] = None

    @classmethod
    def _from_dict(
        cls, config: t.Dict, connection_id: t.Optional[uuid.UUID] = None
    ) -> "Config":
        """

        Parameters
        ----------
        config : t.Dict
            the possible keys to specify inside the config dictionary are:
            REQUIRED:
                - UnicoreApiUrl : str
                    URL to the API of the UNICORE HPC used.

                The other required field MANTIK_USERNAME, MANTIK_PASSWORD
                and MANTIK_PROJECT are inferred from the environment.
            OPTIONAL:
                - Resources : dict
                    Dict of parameters specifying the resources
                    to request on the remote system.
                    More info can be found here:
                    https://sourceforge.net/p/unicore/wiki/Job_Description/
                - Environment : dict
                    Used to build a environment.Environment.
                - Exclude : list
                    List of files or file-patterns
                    that are not sent with the job


        Returns
        -------
        mantik.unicore._config.core.Config

        """
        (
            unicore_api_url,
            firecrest,
            credentials,
        ) = _get_unicore_or_firecrest_and_credentials(
            config=config,
            connection_id=connection_id,
        )
        project = env.get_required_env_var(COMPUTE_BUDGET_ACCOUNT_ENV_VAR)

        # Read remaining config sections.
        resources = _utils.get_required_config_value(
            name="Resources",
            value_type=_resources.Resources.from_dict,
            config=config,
        )
        environment = _utils.get_optional_config_value(
            name="Environment",
            value_type=_environment.Environment.from_dict,
            config=config,
        )
        exclude = _utils.get_optional_config_value(
            name="Exclude",
            value_type=list,
            config=config,
        )

        return cls(
            unicore_api_url=unicore_api_url,
            firecrest=firecrest,
            user=credentials.username,
            password=credentials.password,
            project=project,
            resources=resources,
            environment=environment,
            exclude=exclude,
        )

    def __post_init__(self):
        if self.environment is None:
            self.environment = _environment.Environment()

        # set SRUN_CPUS_PER_TASK to CPUsPerNode if not explicitly set
        self.environment = self.environment.set_srun_cpus_per_task_if_unset(
            self.resources
        )

        if self.firecrest is not None and (
            self.environment.pre_run_command_on_login_node is not None
            or self.environment.post_run_command_on_login_node is not None
        ):
            raise _exceptions.ConfigValidationError(
                "Pre-/PostRunCommandOnLoginNode not supported by firecREST"
            )

    @classmethod
    def from_filepath(
        cls, filepath: pathlib.Path, connection_id: t.Optional[uuid.UUID] = None
    ) -> "Config":
        """Initialize from a given file."""
        return cls._from_dict(
            read.read_config(filepath), connection_id=connection_id
        )

    @property
    def files_to_exclude(self) -> t.List[str]:
        return self.exclude or []

    def _to_dict(self) -> t.Dict:
        environment = (
            self.environment.to_dict() if self.environment is not None else {}
        )
        return {
            "Project": self.project,
            "Resources": self.resources,
            # Write stderr/stdout to given file to allow access to logs
            "Stdout": APPLICATION_LOGS_FILE,
            "Stderr": APPLICATION_LOGS_FILE,
            **environment,
        }

    def to_unicore_job_description(self, bash_script_name: str) -> t.Dict:
        """Convert to UNICORE job description."""
        environment = (
            self.environment.to_unicore_job_description(bash_script_name)
            if self.environment is not None
            else {}
        )
        return {
            **self.to_dict(),
            **environment,
        }

    def to_slurm_batch_script(
        self, arguments: str, run_id: str, run_dir: pathlib.Path
    ) -> str:
        """Create a Slurm batch script for firecREST.

        firecREST only allows to submit jobs via batch scripts, which has to be
        manually constructed.

        """
        return "\n".join(
            [
                "#!/bin/bash -l",
                f"#SBATCH --job-name='mantik-{run_id}'",
                self.resources.to_slurm_batch_script(),
                f"#SBATCH --output={run_dir}/mantik.log",
                f"#SBATCH --error={run_dir}/mantik.log",
                # The `--chdir` option is ignored by firecREST.
                # We could propose this feature, but for now we will
                # stick to manually changing the directory.
                'echo "firecREST working directory is $(pwd)"',
                'echo "Submitted batch script:"',
                "cat $(pwd)/script.batch",
                # _Must_ cd after cat, otherwise the `script.batch` file will
                # not be available.
                f'echo "Changing to Mantik run directory {run_dir}"',
                f"cd {run_dir}",
                self.environment.to_slurm_batch_script(
                    entry_point_arguments=arguments, run_dir=run_dir
                ),
            ]
        )

    def to_bash_script(self, arguments: str) -> str:
        """Create a bash script for UNICORE."""
        return self.environment.to_bash_script(arguments)

    def execution_environment_given(self) -> bool:
        return (
            self.environment is not None and self.environment.execution_given()
        )

    def add_env_vars(self, env_vars: t.Dict) -> None:
        self.environment.add_env_vars(env_vars)

    @property
    def backend_type(self) -> SupportedBackends:
        if self.unicore_api_url is not None:
            return SupportedBackends.UNICORE
        elif self.firecrest is not None:
            return SupportedBackends.FIRECREST
        else:
            raise NotImplementedError()


def _get_unicore_or_firecrest_and_credentials(
    config: t.Dict, connection_id: t.Optional[uuid.UUID] = None
) -> t.Tuple[
    t.Optional[str], t.Optional[_firecrest.Firecrest], _credentials.HpcRestApi
]:
    # Read UNICORE or firecREST info.
    unicore_api_url = _utils.get_optional_config_value(
        name="UnicoreApiUrl",
        value_type=str,
        config=config,
    )
    firecrest = _utils.get_optional_config_value(
        name="Firecrest",
        value_type=_firecrest.Firecrest.from_dict,
        config=config,
    )

    # Check that either UnicoreApiUrl or Firecrest section given
    if unicore_api_url is None and firecrest is None:
        raise _exceptions.ConfigValidationError(
            "Either UnicoreApiUrl or Firecrest details must be provided"
        )
    elif unicore_api_url is not None and firecrest is not None:
        raise _exceptions.ConfigValidationError(
            "Either UnicoreApiUrl or Firecrest details must be provided, "
            "but both are given"
        )

    # Read either UNICORE or firecREST credentials
    if unicore_api_url is not None:
        credentials = _credentials.HpcRestApi.from_unicore_env_vars(
            connection_id
        )
    elif firecrest is not None:
        credentials = _credentials.HpcRestApi.from_firecrest_env_vars(
            connection_id
        )
    else:
        raise RuntimeError("Neither UnicoreApiUrl nor Firecrest section given")

    return unicore_api_url, firecrest, credentials


def backend_type_from_dict(backend_config: t.Dict):
    if "UnicoreApiUrl" in backend_config and "Firecrest" not in backend_config:
        return SupportedBackends.UNICORE
    elif (
        "Firecrest" in backend_config and "UnicoreApiUrl" not in backend_config
    ):
        return SupportedBackends.FIRECREST
    else:
        raise _exceptions.ConfigValidationError(
            "Either UnicoreApiUrl or Firecrest details must be provided"
        )



import pytest

import mantik.utils.credentials as credentials


@pytest.fixture()
def token():
    return (
        "eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9."
        "eyJzdWIiOiIxMjM0NTY3ODkwIiwibmFtZSI6I"
        "kpvaG4gRG9lIiwiaWF0IjoxNTE2MjM5MDIyfQ."
        "SflKxwRJSMeKKF2QT4fwpMeJf36POk6yJV_adQssw5c"
    )


@pytest.fixture()
def sub():
    return "1234567890"


def test_get_sub_from_token(token, sub):
    assert credentials._get_sub_from_token(token) == sub



"""MLflow-related util functions.

This source file contains code from [MLflow](https://github.com/mlflow/mlflow)
licensed under Apache-2.0 license, see
[here](https://github.com/mlflow/mlflow/blob/1eef4641df6f605b7d9faa83b0fc25e65877dbf4/LICENSE.txt)  # noqa: E501
for the original license.

Changes made to the original source code are denoted as such with comments.
"""
import mantik.utils.env as env

TRACKING_URI_ENV_VAR = "MLFLOW_TRACKING_URI"
TRACKING_TOKEN_ENV_VAR = "MLFLOW_TRACKING_TOKEN"
TRACKING_USERNAME_ENV_VAR = "MLFLOW_TRACKING_USERNAME"
TRACKING_PASSWORD_ENV_VAR = "MLFLOW_TRACKING_PASSWORD"
EXPERIMENT_NAME_ENV_VAR = "MLFLOW_EXPERIMENT_NAME"
EXPERIMENT_ID_ENV_VAR = "MLFLOW_EXPERIMENT_ID"
ACTIVE_RUN_ID_ENV_VAR = "ACTIVE_RUN_ID"

CONFLICTING_ENV_VARS = (
    TRACKING_USERNAME_ENV_VAR,
    TRACKING_PASSWORD_ENV_VAR,
)

DEFAULT_TRACKING_URI = "https://api.cloud.mantik.ai/tracking/"


def unset_conflicting_env_vars() -> None:
    env.unset_env_vars(CONFLICTING_ENV_VARS)



import dataclasses
import typing as t

import mantik.config._base as _base
import mantik.config._utils as _utils


@dataclasses.dataclass
class Firecrest(_base.ConfigObject):
    api_url: str
    token_url: str
    machine: str

    @classmethod
    def _from_dict(cls, config: t.Dict) -> "Firecrest":
        api_url = _utils.get_required_config_value(
            name="ApiUrl",
            value_type=str,
            config=config,
        )
        token_url = _utils.get_required_config_value(
            name="TokenUrl",
            value_type=str,
            config=config,
        )
        machine = _utils.get_required_config_value(
            name="Machine",
            value_type=str,
            config=config,
        )
        return cls(api_url=api_url, token_url=token_url, machine=machine)

    def _to_dict(self) -> t.Dict:
        return {
            "ApiUrl": self.api_url,
            "TokenUrl": self.token_url,
            "Machine": self.machine,
        }



import datetime
import os
import typing as t

import pytest

import mantik
import mantik.authentication.api as _api
import mantik.authentication.auth as auth
import mantik.authentication.tokens as _tokens
import mantik.testing.token as testing_token
import mantik.tracking.track as track
import mantik.utils.mantik_api.client as mantik_api


@pytest.fixture(autouse=True)
def set_required_env_vars(
    env_vars_set, mantik_api_url, expect_raise_if_exception
):
    env_vars = {mantik_api._MANTIK_API_URL_ENV_VAR: mantik_api_url}
    with env_vars_set(env_vars):
        yield


def test_track_without_existing_tokens(
    requests_mock,
    mantik_api_url,
    tmp_dir_as_test_mantik_folder,
    required_env_vars,
    token_expiration_date,
):
    requests_mock.post(
        url=f"{mantik_api_url}"
        f"{mantik_api.MANTIK_API_CREATE_TOKEN_API_PATH}",
        json={
            "AccessToken": "test-access-token",
            "RefreshToken": "test-refresh-token",
            "ExpiresAt": token_expiration_date.isoformat(),
        },
    )

    _init_tracking_and_assert_expected(
        env_vars=required_env_vars,
        expected_access_token="test-access-token",
    )
    # Cleanup
    os.unsetenv(mantik.utils.mlflow.TRACKING_TOKEN_ENV_VAR)


@testing_token.set_token(
    access_token="test-stored-access-token",
    refresh_token="test-stored-refresh-token",
)
def test_track_with_existing_valid_tokens(
    monkeypatch,
    tmp_dir_as_test_mantik_folder,
    required_env_vars,
    token_expiration_date,
):
    _init_tracking_and_assert_expected(
        env_vars=required_env_vars,
        expected_access_token="test-stored-access-token",
    )


@testing_token.set_token(
    access_token="test-stored-expired-access-token",
    refresh_token="test-stored-refresh-token",
    expires_at=datetime.datetime(2020, 1, 1),
)
def test_track_with_existing_expired_token(
    requests_mock,
    mantik_api_url,
    tmp_dir_as_test_mantik_folder,
    required_env_vars,
    token_expiration_date,
):
    requests_mock.post(
        url=f"{mantik_api_url}"
        f"{mantik_api.MANTIK_API_REFRESH_TOKEN_API_PATH}",
        json={
            "AccessToken": "test-refreshed-access-token",
            "ExpiresAt": token_expiration_date.isoformat(),
        },
    )

    _init_tracking_and_assert_expected(
        env_vars=required_env_vars,
        expected_access_token="test-refreshed-access-token",
    )


@testing_token.set_token(
    access_token="test-stored-expired-access-token", refresh_token="test-stored"
)
def test_track_with_existing_expired_token_and_expired_refresh_token(
    requests_mock,
    mantik_api_url,
    tmp_dir_as_test_mantik_folder,
    required_env_vars,
    token_expiration_date,
):
    # Mock refresh api to return expired error
    requests_mock.post(
        url=f"{mantik_api_url}"
        f"{mantik_api.MANTIK_API_REFRESH_TOKEN_API_PATH}",
        status_code=401,
        text=_api.REFRESH_TOKEN_EXPIRED_ERROR_MESSAGE,
    )
    # Mock get api to return refreshed tokens
    requests_mock.post(
        url=(
            f"{mantik_api_url}" f"{mantik_api.MANTIK_API_CREATE_TOKEN_API_PATH}"
        ),
        json={
            "AccessToken": "test-refreshed-access-token",
            "RefreshToken": "test-refreshed-refresh-token",
            "ExpiresAt": token_expiration_date.isoformat(),
        },
    )

    tokens = _tokens.Tokens(
        access_token="test-stored-expired-access-token",
        refresh_token="test-stored-refresh-token",
        expires_at=datetime.datetime(2020, 1, 1),
    )
    tokens.write_to_file(auth._MANTIK_TOKEN_FILE)

    _init_tracking_and_assert_expected(
        env_vars=required_env_vars,
        expected_access_token="test-refreshed-access-token",
    )

    result_tokens = _tokens.Tokens.from_file(auth._MANTIK_TOKEN_FILE)
    assert result_tokens.access_token == "test-refreshed-access-token"
    assert result_tokens.refresh_token == "test-refreshed-refresh-token"
    assert result_tokens.expires_at == token_expiration_date


@testing_token.set_token(
    access_token="test-stored-expired-access-token",
    refresh_token="test-stored-invalid-refresh-token",
    expires_at=datetime.datetime(2020, 1, 1),
)
def test_track_with_existing_invalid_refresh_token(
    requests_mock,
    mantik_api_url,
    tmp_dir_as_test_mantik_folder,
    required_env_vars,
    token_expiration_date,
):
    # Mock refresh api to return invalid error
    requests_mock.post(
        url=f"{mantik_api_url}"
        f"{mantik_api.MANTIK_API_REFRESH_TOKEN_API_PATH}",
        status_code=401,
        text=_api.REFRESH_TOKEN_INVALID_ERROR_MESSAGE,
    )
    # Mock get api to return refreshed tokens
    requests_mock.post(
        url=(
            f"{mantik_api_url}" f"{mantik_api.MANTIK_API_CREATE_TOKEN_API_PATH}"
        ),
        json={
            "AccessToken": "test-refreshed-access-token",
            "RefreshToken": "test-refreshed-refresh-token",
            "ExpiresAt": token_expiration_date.isoformat(),
        },
    )

    _init_tracking_and_assert_expected(
        env_vars=required_env_vars,
        expected_access_token="test-refreshed-access-token",
    )

    result_tokens = _tokens.Tokens.from_file(auth._MANTIK_TOKEN_FILE)
    assert result_tokens.access_token == "test-refreshed-access-token"
    assert result_tokens.refresh_token == "test-refreshed-refresh-token"
    assert result_tokens.expires_at == token_expiration_date


def _init_tracking_and_assert_expected(
    env_vars: t.Dict[str, str],
    expected_access_token: str,
) -> None:
    with mantik.utils.env.env_vars_set(env_vars):
        track.init_tracking()

        mantik.testing.env.assert_conflicting_mlflow_env_vars_not_set()
        mantik.testing.env.assert_correct_tracking_token_env_var_set(
            expected_access_token
        )



import datetime

import pytest

import mantik.remote_file_service.abstract_file_service as afs
import mantik.remote_file_service.data_client as _data_client


class TestDataClient:
    def test_list_directory(self, data_client, fake_unicore_fs):
        target = "test_target"
        assert data_client.list_directory(
            target
        ) == fake_unicore_fs.list_directory(target)

    def test_create_directory(self, data_client, fake_unicore_fs, project_id):
        target = "test_target"
        assert data_client.create_directory(
            target,
        ) == fake_unicore_fs.create_directory(target)

    def test_remove_directory(self, data_client, fake_unicore_fs, project_id):
        target = "test_target"
        assert data_client.remove_directory(
            target
        ) == fake_unicore_fs.remove_directory(target)

    def test_copy_directory(self, data_client, fake_unicore_fs, project_id):
        source = "test_source"
        target = "test_target"
        assert data_client.copy_directory(
            source,
            target,
        ) == fake_unicore_fs.copy_directory(source, target)

    def test_create_file_if_not_exists(
        self, data_client, fake_unicore_fs, project_id
    ):
        target = "test_target"
        assert data_client.create_file_if_not_exists(
            target,
        ) == fake_unicore_fs.create_file_if_not_exists(target)

    def test_remove_file(self, data_client, fake_unicore_fs, project_id):
        target = "test_target"
        assert data_client.remove_file(target) == fake_unicore_fs.remove_file(
            target
        )

    def test_copy_file(self, data_client, fake_unicore_fs, project_id):
        source = "test_source"
        target = "test_target"
        assert data_client.copy_file(
            source,
            target,
        ) == fake_unicore_fs.copy_file(source, target)

    def test_user(self, data_client, fake_unicore_fs):
        assert data_client.user == fake_unicore_fs.user

    def test_change_permissions(self, data_client, fake_unicore_fs):
        target = "test_target"
        new_permissions = afs.FileMeta(
            last_changed=datetime.datetime(2022, 1, 1),
            mode="xxx",
            owner=fake_unicore_fs.user,
        )
        assert data_client.change_permissions(
            target, new_permissions
        ) == fake_unicore_fs.change_permissions(target, new_permissions)

    def test_init_not_supported_backend(self, env_vars_set):
        with env_vars_set(
            {
                "MANTIK_PROJECT_ID": "3fa85f64-5717-4562-b3fc-2c963f66afa6",
                "REMOTE_FS_TYPE": "NEW_BACKEND",
            }
        ), pytest.raises(_data_client.DataClientException) as e:
            assert _data_client.DataClient.from_env()

        expected = (
            "Invalid remote file system type, "
            "set REMOTE_FS_TYPEas one of this supported "
            "types ['S3', 'UNICORE']"
        )

        result = str(e.value)

        assert result == expected



import pathlib

import pytest

FILE_DIR = pathlib.Path(__file__).parent


@pytest.fixture()
def example_project_absolute_path() -> str:
    return (FILE_DIR / "../../../resources/test-project").as_posix()


@pytest.fixture()
def example_project_relative_path() -> str:
    return "../test-project"



import os
import pathlib
import unittest.mock
import uuid

import pytest
import responses

import mantik.cli.data
import mantik.data_repository.data_repository
import mantik.utils.env_vars as env_vars

FAKE_API_URL = "https://fake.com"


@pytest.fixture(scope="session")
def sample_data_repository_uuid():
    return uuid.uuid4()


@pytest.fixture(scope="session")
def sample_target_dir() -> pathlib.Path:
    return pathlib.Path("/sample/target/dir")


@pytest.fixture(scope="function")
def sample_version() -> str:
    return "1234"


@pytest.fixture
def mock_data_repository_response_with_dvc(
    sample_project_uuid,
    sample_data_repository_uuid,
    sample_version,
    sample_connection_uuid,
    mock_mantik_api_url,
):
    responses.add(
        responses.GET,
        url=f"{FAKE_API_URL}/projects/{str(sample_project_uuid)}/data"
        f"/{str(sample_data_repository_uuid)}",
        status=200,
        json={
            "uri": "fake.com",
            "isDvcEnabled": True,
            "dvcConnectionId": str(sample_connection_uuid),
            "versions": {sample_version: "fake-commit-hash"},
            "platform": "GitLab",
        },
    )
    yield


@pytest.fixture
def mock_data_repository_response_without_dvc(
    sample_project_uuid,
    sample_data_repository_uuid,
    sample_version,
    mock_mantik_api_url,
):
    responses.add(
        responses.GET,
        url=f"{FAKE_API_URL}/projects/{str(sample_project_uuid)}/data"
        f"/{str(sample_data_repository_uuid)}",
        status=200,
        json={
            "uri": "fake.com",
            "isDvcEnabled": False,
            "dvcConnectionId": None,
            "versions": {sample_version: "fake-commit-hash"},
            "platform": "GitLab",
        },
    )
    yield


@pytest.fixture
def mock_connection_response(
    sample_project_uuid,
    fake_user_id_str,
    sample_connection_uuid,
    mock_mantik_api_url,
):
    responses.add(
        responses.GET,
        url=f"{FAKE_API_URL}/users/{fake_user_id_str}/settings/"
        f"connections/{str(sample_connection_uuid)}",
        status=200,
        json={
            "connectionId": str(sample_connection_uuid),
            "user": {"userId": fake_user_id_str, "name": "Kiwi"},
            "connectionName": "KiwiConnection",
            "connectionProvider": "S3",
            "authMethod": "Username-Password",
            "loginName": "access-key",
            "password": "secret-access-key",
            "token": "token",
        },
    )
    yield


@responses.activate
def test_download_data_repository_without_dvc(
    sample_project_uuid,
    sample_data_repository_uuid,
    sample_target_dir,
    sample_version,
    mock_data_repository_response_without_dvc,
):
    with unittest.mock.patch(
        "mantik.data_repository.data_repository.git_clone_with_checkout"
    ) as mocked_git_clone_function, unittest.mock.patch(
        "mantik.data_repository.data_repository.dvc_pull_with_aws_credentials"
    ) as mocked_dvc_pull_with_aws_credentials_function:
        mantik.data_repository.data_repository.download_data_repository(
            project_id=sample_project_uuid,
            data_repository_id=sample_data_repository_uuid,
            checkout=sample_version,
            target_dir=sample_target_dir,
            token="1234",
        )

        mocked_git_clone_function.assert_called()
        mocked_dvc_pull_with_aws_credentials_function.assert_not_called()


@responses.activate
def test_download_data_repository_with_dvc(
    sample_project_uuid,
    sample_data_repository_uuid,
    sample_target_dir,
    sample_version,
    mock_connection_response,
    mock_data_repository_response_with_dvc,
    mock_get_user_id_from_token,
):
    with unittest.mock.patch(
        "mantik.data_repository.data_repository.git_clone_with_checkout"
    ) as mocked_git_clone_function, unittest.mock.patch(
        "mantik.data_repository.data_repository.dvc_pull_with_aws_credentials"
    ) as mocked_dvc_pull_with_aws_credentials_function:
        mantik.data_repository.data_repository.download_data_repository(
            project_id=sample_project_uuid,
            data_repository_id=sample_data_repository_uuid,
            checkout=sample_version,
            target_dir=sample_target_dir,
            token="1234",
        )

        mocked_git_clone_function.assert_called()
        mocked_dvc_pull_with_aws_credentials_function.assert_called()


@pytest.fixture()
def mock_env():
    os.environ[env_vars.PROJECT_ID_ENV_VAR] = str(uuid.uuid4())
    os.environ[env_vars.DATA_REPOSITORY_ID_ENV_VAR] = str(uuid.uuid4())
    os.environ[env_vars.TARGET_DIR_ENV_VAR] = "."
    os.environ[env_vars.MANTIK_ACCESS_TOKEN_ENV_VAR] = "1234"

    yield

    os.unsetenv(env_vars.PROJECT_ID_ENV_VAR)
    os.unsetenv(env_vars.DATA_REPOSITORY_ID_ENV_VAR)
    os.unsetenv(env_vars.TARGET_DIR_ENV_VAR)
    os.unsetenv(env_vars.MANTIK_ACCESS_TOKEN_ENV_VAR)


def test_data_download_fails_if_no_environment_variables() -> None:
    with pytest.raises(RuntimeError):
        with unittest.mock.patch(
            "mantik.data_repository.data_repository.download_data_repository"
        ) as patch:
            mantik.data_download()
            patch.assert_not_called()


def test_data_download_works_without_explicitly_specifying_arguments(
    mock_env,
) -> None:
    with unittest.mock.patch(
        "mantik.data_repository.data_repository.download_data_repository"
    ) as patch:
        mantik.data_download()
        patch.assert_called()



import dataclasses
import typing as t

import pytest

import mantik.config._base as _base


@dataclasses.dataclass
class AnyConfigObject(_base.ConfigObject):
    test: str = "test"

    @classmethod
    def _from_dict(cls, config: t.Dict) -> "AnyConfigObject":
        pass

    def _to_dict(self) -> t.Dict:
        return dataclasses.asdict(self)


@pytest.mark.parametrize(
    ("dict_", "expected"),
    [
        (
            {"any": "value"},
            {"any": "value"},
        ),
        (
            {"any": None},
            {},
        ),
        (
            {"any": AnyConfigObject()},
            {"any": {"test": "test"}},
        ),
    ],
)
def test_exclude_none_values(dict_, expected):
    result = _base._exclude_none_values(dict_)

    assert result == expected



import pathlib

import mantik.config
import mantik.config.environment as environment
import mantik.config.read as read


def test_read_config(compute_backend_config_yaml):
    backend_config = read.read_config(compute_backend_config_yaml)
    assert (
        backend_config["UnicoreApiUrl"]
        == "https://zam2125.zam.kfa-juelich.de:9112/JUWELS/rest/core"
    )
    assert backend_config["Resources"]["Queue"] == "devel"
    assert backend_config["Resources"]["Nodes"] == 2


def test_read_config_file_not_fund(
    expect_raise_if_exception, compute_backend_config_yaml
):
    with expect_raise_if_exception(
        mantik.config.exceptions.ConfigValidationError()
    ) as e:
        read.read_config(pathlib.Path("/does/not/exist.yaml"))

        assert "No such config: exist.yaml" in str(e)


def test_read_config_unsupported_type(expect_raise_if_exception):
    unsupported_format = ".yamml"
    with expect_raise_if_exception(
        mantik.config.exceptions.ConfigValidationError()
    ) as e:
        read.read_config(pathlib.Path(f"backend-config{unsupported_format}"))
    assert (
        "The given file type '.yamml' is not supported for the config, "
        "the supported ones are: '.json', '.yml', '.yaml'."
    ) == str(e.value)


def test_read_yaml_config(compute_backend_config_yaml):
    backend_config = read._read_yaml_config(compute_backend_config_yaml)
    assert (
        backend_config["UnicoreApiUrl"]
        == "https://zam2125.zam.kfa-juelich.de:9112/JUWELS/rest/core"
    )
    assert backend_config["Resources"]["Queue"] == "devel"
    assert backend_config["Resources"]["Nodes"] == 2


def test_read_json_config(compute_backend_config_json):
    backend_config = read._read_yaml_config(compute_backend_config_json)
    assert (
        backend_config["UnicoreApiUrl"]
        == "https://zam2125.zam.kfa-juelich.de:9112/JUWELS/rest/core"
    )
    assert backend_config["Resources"]["Queue"] == "devel"
    assert backend_config["Resources"]["Nodes"] == 2
    assert backend_config["Environment"]["Apptainer"] == {
        "Path": "mantik-test.sif",
        "Type": "lOcAl",
    }


def test_read_run_commands_from_config(compute_backend_config_yaml):
    backend_config = read._read_yaml_config(compute_backend_config_yaml)
    assert backend_config["Environment"]["PreRunCommandOnComputeNode"] == [
        "echo precommand",
        "echo on compute node",
        1,
    ]
    assert backend_config["Environment"]["PostRunCommandOnComputeNode"] == [
        "echo postcommand",
        "echo on compute node",
    ]
    assert backend_config["Environment"]["PreRunCommandOnLoginNode"] == [
        "echo precommand",
        "echo on login node",
    ]
    assert (
        backend_config["Environment"]["PostRunCommandOnLoginNode"]
        == "echo postcommand echo on login node"
    )

    env = environment.Environment.from_dict(
        config=backend_config["Environment"]
    )

    assert env.pre_run_command_on_compute_node == [
        "echo precommand",
        "echo on compute node",
        "1",
    ]
    assert env.post_run_command_on_compute_node == [
        "echo postcommand",
        "echo on compute node",
    ]
    assert env.pre_run_command_on_login_node == [
        "echo precommand",
        "echo on login node",
    ]
    assert env.post_run_command_on_compute_node == [
        "echo postcommand",
        "echo on compute node",
    ]



import functools
import json
import pathlib
import uuid

import firecrest as firecrest
import mlflow.projects as projects

import mantik.config.core as core
import mantik.testing as testing
import mantik.utils as mantik_utils
import mantik.utils.credentials as _credentials
import mantik_compute_backend.backend as backend

FILE_PATH = pathlib.Path(__file__).parent

ALL_ENV_VARS = [
    _credentials.FIRECREST_CLIENT_ID_ENV_VAR,
    _credentials.FIRECREST_CLIENT_SECRET_ENV_VAR,
    core.COMPUTE_BUDGET_ACCOUNT_ENV_VAR,
]


class FakeProject:
    pass


class TestFirecrestBackend:
    def test_run(
        self, monkeypatch, example_project_path, tmp_path, env_vars_set, caplog
    ):
        monkeypatch.setattr(
            firecrest,
            "Firecrest",
            testing.firecrest.FakeClient,
        )
        monkeypatch.setattr(
            firecrest,
            "ClientCredentialsAuth",
            functools.partial(
                testing.firecrest.FakeClientCredentialsAuth,
                login_successful=True,
            ),
        )

        backend_config_path = (
            example_project_path / "compute-backend-firecrest-config.json"
        )
        with open(backend_config_path) as infile:
            backend_config = json.load(infile)
        # Following env vars are set by MLflow before running a project.
        backend_config[
            projects.PROJECT_STORAGE_DIR
        ] = example_project_path.as_posix()

        # Point MLFLOW_TRACKING_URI to a temporary directory
        tracking_uri = tmp_path / "mlruns"

        with env_vars_set(
            {
                # Following env vars must be set for the config.
                **{key: "test-val" for key in ALL_ENV_VARS},
                mantik_utils.mlflow.TRACKING_URI_ENV_VAR: tracking_uri.as_uri(),
                mantik_utils.mlflow.ACTIVE_RUN_ID_ENV_VAR: str(
                    uuid.uuid4().hex
                ),
            }
        ):
            backend.Backend().run(
                project_uri=example_project_path.as_posix(),
                entry_point="main",
                params={"print": "test"},
                version=None,
                backend_config=backend_config,
                tracking_uri=None,
                experiment_id=None,
            )
            assert "Submitted job test-job-id via firecrest" in caplog.text


def test_create_job_script_apptainer(
    example_project_path, example_project, example_firecrest_config
):
    entry_point = example_project.get_entry_point("main")
    parameters = {"print": "whatever"}
    storage_dir = "test-dir"
    run_id = str(uuid.uuid4())
    run_dir = pathlib.Path("test-dir")
    expected = "\n".join(
        [
            "#!/bin/bash -l",
            f"#SBATCH --job-name='mantik-{run_id}'",
            "#SBATCH --partition=batch",
            "#SBATCH --constraint=gpu",
            f"#SBATCH --output={run_dir}/mantik.log",
            f"#SBATCH --error={run_dir}/mantik.log",
            'echo "firecREST working directory is $(pwd)"',
            'echo "Submitted batch script:"',
            "cat $(pwd)/script.batch",
            'echo "Changing to Mantik run directory test-dir"',
            f"cd {run_dir}",
            "export MANTIK_WORKING_DIRECTORY=test-dir",
            "srun apptainer run mantik-test.sif python main.py whatever",
        ]
    )

    arguments = backend._create_arguments(
        entry_point=entry_point, parameters=parameters, storage_dir=storage_dir
    )
    result = example_firecrest_config.to_slurm_batch_script(
        arguments=arguments,
        run_id=run_id,
        run_dir=run_dir,
    )

    assert result == expected


def test_create_job_script_apptainer_long_resources(
    example_project_path,
    example_project,
    example_firecrest_config_long_resources,
):
    entry_point = example_project.get_entry_point("main")
    parameters = {"print": "whatever"}
    storage_dir = "test-dir"
    run_id = str(uuid.uuid4())
    run_dir = pathlib.Path("test-dir")
    expected = "\n".join(
        [
            "#!/bin/bash -l",
            f"#SBATCH --job-name='mantik-{run_id}'",
            "#SBATCH --partition=batch",
            "#SBATCH --time=0-01:00:00",
            "#SBATCH --nodes=1",
            "#SBATCH --ntasks=1",
            "#SBATCH --ntasks-per-node=1",
            "#SBATCH --gpus-per-node=1",
            "#SBATCH --mem=1",
            "#SBATCH --reservation=test-reservation",
            "#SBATCH --constraint=gpu",
            "#SBATCH --qos=test-qos,",
            "#SBATCH --output=test-dir/mantik.log",
            "#SBATCH --error=test-dir/mantik.log",
            'echo "firecREST working directory is $(pwd)"',
            'echo "Submitted batch script:"',
            "cat $(pwd)/script.batch",
            'echo "Changing to Mantik run directory test-dir"',
            f"cd {run_dir}",
            "export MANTIK_WORKING_DIRECTORY=test-dir",
            "export SRUN_CPUS_PER_TASK=1",
            "srun apptainer run mantik-test.sif python main.py whatever",
        ]
    )

    arguments = backend._create_arguments(
        entry_point=entry_point, parameters=parameters, storage_dir=storage_dir
    )
    result = example_firecrest_config_long_resources.to_slurm_batch_script(
        arguments=arguments,
        run_id=run_id,
        run_dir=run_dir,
    )

    assert result == expected


def test_create_job_script_apptainer_entry_point_multi_line(
    example_project_path, example_project, example_firecrest_config
):
    entry_point = example_project.get_entry_point("multi-line")
    parameters = {"print": "whatever"}
    storage_dir = "test-dir"
    run_id = str(uuid.uuid4())
    run_dir = pathlib.Path("test-dir")
    expected = "\n".join(
        [
            "#!/bin/bash -l",
            f"#SBATCH --job-name='mantik-{run_id}'",
            "#SBATCH --partition=batch",
            "#SBATCH --constraint=gpu",
            f"#SBATCH --output={run_dir}/mantik.log",
            f"#SBATCH --error={run_dir}/mantik.log",
            'echo "firecREST working directory is $(pwd)"',
            'echo "Submitted batch script:"',
            "cat $(pwd)/script.batch",
            'echo "Changing to Mantik run directory test-dir"',
            f"cd {run_dir}",
            "export MANTIK_WORKING_DIRECTORY=test-dir",
            (
                "srun apptainer run mantik-test.sif python main.py whatever  "
                "--o option1  --i option2"
            ),
        ]
    )

    arguments = backend._create_arguments(
        entry_point=entry_point, parameters=parameters, storage_dir=storage_dir
    )
    result = example_firecrest_config.to_slurm_batch_script(
        arguments=arguments,
        run_id=run_id,
        run_dir=run_dir,
    )
    assert result == expected


def test_create_job_description_python(
    example_project_path, example_project, example_firecrest_config_for_python
):
    entry_point = example_project.get_entry_point("main")
    parameters = {"print": "whatever"}
    storage_dir = "test-dir"
    run_id = str(uuid.uuid4())
    run_dir = pathlib.Path("test-dir")
    expected = "\n".join(
        [
            "#!/bin/bash -l",
            f"#SBATCH --job-name='mantik-{run_id}'",
            "#SBATCH --partition=batch",
            "#SBATCH --constraint=gpu",
            f"#SBATCH --output={run_dir}/mantik.log",
            f"#SBATCH --error={run_dir}/mantik.log",
            'echo "firecREST working directory is $(pwd)"',
            'echo "Submitted batch script:"',
            "cat $(pwd)/script.batch",
            'echo "Changing to Mantik run directory test-dir"',
            f"cd {run_dir}",
            "export MANTIK_WORKING_DIRECTORY=test-dir",
            "precommand compute node && source /venv/bin/activate "
            "&& python main.py "
            "whatever && postcommand compute node",
        ]
    )

    arguments = backend._create_arguments(
        entry_point=entry_point, parameters=parameters, storage_dir=storage_dir
    )
    result = example_firecrest_config_for_python.to_slurm_batch_script(
        arguments=arguments,
        run_id=run_id,
        run_dir=run_dir,
    )
    assert result == expected



import os
import pathlib

import pytest

FILE_DIR = pathlib.Path(__file__).parent


@pytest.fixture()
def unicore_api_url() -> str:
    # Allow reading from env var for CI testing
    return os.getenv(
        "UNICORE_API_TESTING_URL", "https://localhost:8080/DEMO-SITE/rest/core"
    )


@pytest.fixture()
def unicore_credentials() -> dict:
    # Username and password are provided in
    # https://github.com/UNICORE-EU/tools/blob/master/unicore-docker-image/README.md
    return {
        "user": "demouser",
        "password": "test123",
    }


@pytest.fixture()
def example_project_with_batch_script_path() -> pathlib.Path:
    return FILE_DIR / "../../resources/test-project-with-batch-script"
